#!/usr/bin/env python
# FILE: visualizeClassification.py
# AUTHOR: Duong Vu
# CREATE DATE: 07 June 2021
#from sklearn.model_selection import StratifiedKFold
#from sklearn.preprocessing import StandardScaler
import sys
if sys.version_info[0] >= 3:
	unicode = str
import os, argparse
from sklearn.metrics import precision_recall_fscore_support
from sklearn.metrics import cohen_kappa_score
from sklearn.metrics import matthews_corrcoef
from sklearn.metrics import confusion_matrix
from sklearn.metrics import accuracy_score
import json
from Bio import SeqIO
import random

parser=argparse.ArgumentParser(prog='visualizeClassification.py',  
							   usage="%(prog)s [options] -i assignmentfile -c classification -o kronareport",
							   description='''Script that visualizes using Krona for the classification results.''',
							   epilog="""Written by Duong Vu duong.t.vu@gmail.com""",
   )

parser.add_argument('-i','--input', required=True, help='the assignment file')
parser.add_argument('-o','--out', default="dnabarcoder", help='The output folder.')
parser.add_argument('-c','--classification', default ="", help='the classification file of the reference sequences.')

args=parser.parse_args()
predictionfilename=args.input
classificationfilename=args.classification
outputpath=args.out
if not os.path.exists(outputpath):
	os.system("mkdir " + outputpath)

def GetBase(filename):
	return filename[:-(len(filename)-filename.rindex("."))]
	
def GetWorkingBase(filename):
	basename=os.path.basename(filename)
	if "." in basename:
		basename=basename[:-(len(basename)-basename.rindex("."))] 
	path=outputpath + "/" + filename
	return path

def GetTaxonomicClassification(level,header,texts):
	classification=""
	p_s=len(texts)
	p_g=len(texts)
	p_f=len(texts)
	p_o=len(texts)
	p_c=len(texts)
	p_p=len(texts)
	p_k=len(texts)
	i=0
	for text in header.split("\t"):
		text=text.rstrip()
		if text.lower()=="species":
			p_s=i
		elif text.lower()=="genus":
			p_g=i	
		elif text.lower()=="family":
			p_f=i	
		elif text.lower()=="order":
			p_o=i	
		elif text.lower()=="class":
			p_c=i	
		elif text.lower()=="phylum":
			p_p=i	
		elif text.lower()=="kingdom":
			p_k=i	
		i=i+1 
	species="unidentified"
	genus="unidentified"
	family="unidentified"
	order="unidentified" 
	bioclass="unidentified"
	phylum="unidentified"
	kingdom="unidentified"
	if p_s< len(texts):
		species=texts[p_s].rstrip()
	if p_g< len(texts):
		genus=texts[p_g].rstrip()
	if p_f< len(texts):
		family=texts[p_f].rstrip()
	if p_o< len(texts):
		order=texts[p_o].rstrip()
	if p_c< len(texts):
		bioclass=texts[p_c].rstrip()
	if p_p< len(texts):
		phylum=texts[p_p].rstrip()
	if p_k< len(texts):
		kingdom=texts[p_k].rstrip()
	taxonname=""
	rank=""
	if level <7 and kingdom!="unidentified":
		taxonname=kingdom
		rank="kingdom"
		classification="k__" + kingdom +";p__unidentified;c__unidentified;o__unidentified;f__unidentified;g__unidentified;s__unidentified"	
	if level <6 and phylum!="unidentified":
		taxonname=phylum
		rank="phylum"
		classification="k__" + kingdom +";p__"+phylum +";c__unidentified;o__unidentified;f__unidentified;g__unidentified;s__unidentified"
	if level <5 and bioclass!="unidentified":
		taxonname=bioclass
		rank="class"
		classification="k__" + kingdom +";p__"+phylum +";c__"+bioclass+";o__unidentified;f__unidentified;g__unidentified;s__unidentified"
	if level <4 and order!="unidentified":
		taxonname=order
		rank="order"
		classification="k__" + kingdom +";p__"+phylum +";c__"+bioclass +";o__"+ order + ";f__unidentified;g__unidentified;s__unidentified"
	if level <3 and family!="unidentified":
		taxonname=family
		rank="family"
		classification="k__" + kingdom +";p__"+phylum +";c__"+bioclass +";o__"+ order+";f__"+family +";g__unidentified;s__unidentified"
	if level <2 and genus!="unidentified":
		taxonname=genus
		rank="genus"
		classification="k__" + kingdom +";p__"+phylum +";c__"+bioclass +";o__"+ order+";f__"+family + ";g__"+ genus + "s__unidentified"
	if level <1 and species!="unidentified":
		taxonname=species
		rank="species"
		classification="k__" + kingdom +";p__"+phylum +";c__"+bioclass +";o__"+ order+";f__"+family + ";g__"+ genus+";s__"+species 	
	return classification,taxonname,rank

def LoadClassification(classificationfilename):
	classificationdict={}
	if not os.path.exists(classificationfilename):
		return classificationdict
	classificationfile=open(classificationfilename)
	header=next(classificationfile)
	for line in classificationfile:
		texts=line.split("\t")
		seqid = texts[0].replace(">","").rstrip()
		classificationdict.setdefault(seqid,"")
		classification,taxonname,rank=GetTaxonomicClassification(0,header,texts)
		classificationdict[seqid]=classification
	classificationfile.close()	
	return classificationdict

def GetLevel(rank):
	level=-1
	if rank=="species":	
		level=6
	elif rank=="genus":	
		level=5
	elif rank=="family":	
		level=4
	elif rank=="order":	
		level=3
	elif rank=="class":	
		level=2
	elif rank=="phylum":
		level=1
	elif rank=="kingdom":
		level=0	
	return level

def LoadPrediction(predictionfilename,classificationdict):
	classificationdict={}
	predictionfile= open(predictionfilename, "r")
	next(predictionfile)
	for line in predictionfile:
		texts=line.split("\t")
		classification=texts[3]
		classification=classification.replace("k__","").replace("p__","").replace("c__","").replace("o__","").replace("f__","").replace("g__","").replace("s__","")
		if ";" in classification:
			texts=classification.split(";")
			classification=texts[0]
			if classification=="":
				classification="unidentified"
			for i in range(1,len(texts)):
				text=texts[i]
				if text=="":
					text="unidentified"
				classification=classification + "\t" + text	
		if classification in classificationdict.keys():
			classificationdict[classification]=classificationdict[classification] + 1
		else:
			classificationdict.setdefault(classification, 1)	
	return classificationdict

def KronaPieCharts(classification,kronareport,kronahtml):
	kronareportfile=open(kronareport,"w")
	for classname in classification.keys():
		kronareportfile.write(str(classification[classname]) + "\t" + classname + "\n")
	kronareportfile.close()	
	#create kronahtml
	command="ImportText.pl " + kronareport + " -o " + kronahtml
	#print(command)
	os.system(command)
	os.system("firefox " + kronahtml) 
	
if __name__ == "__main__":
	classificationdict=LoadClassification(classificationfilename)
	classificationdict= LoadPrediction(predictionfilename,classificationdict)
	#making krona report
	kronareport = GetWorkingBase(predictionfilename) + ".krona.report"
	kronahtml=GetWorkingBase(kronareport) + ".html"
	KronaPieCharts(classificationdict,kronareport,kronahtml)
	print("The krona report and html are saved in files " + kronareport + " and " + kronahtml + ".") 



#!/usr/bin/env python
# FILE: select58Ssequences.py, requiring Python 2.7 if classification file contains non-ascii characters
# AUTHOR: Duong Vu
# CREATE DATE: 07 June 2021
import sys, argparse
import numpy as np
import os
from Bio import SeqIO
import json
import multiprocessing
nproc=multiprocessing.cpu_count()
#from keras.utils import np_utils

parser=argparse.ArgumentParser(prog='select58Ssequences.py',  
							   usage="%(prog)s [options] -i fastafile  -t ITSxpositionfile -o output",
							   description='''Script that select the 5.8S regions from the sequences. ''',
							   epilog="""Written by Duong Vu duong.t.vu@gmail.com""",
   )

parser.add_argument('-i','--input', required=True, help='the fasta file.')
parser.add_argument('-t','--ITSxpositions', required=True, help='the ITSx positions file obtained from the ITSx software.')
parser.add_argument('-o','--out', help='The output fasta file with 5.8S sequences.')

args=parser.parse_args()
fastafilename= args.input
ITSxpositionfilename=args.ITSxpositions
output=args.out

ITSxpositionfile=open(ITSxpositionfilename)
positiondict={}
for line in ITSxpositionfile:
	texts=line.split("\t")
	seqid=texts[0]
	if not "5.8S: " in line:
		continue
	texts=line.split("5.8S: ")
	text=texts[1]
	text=text.split("\t")[0]
	if not "-" in text:
		continue
	startpos=int(text.split("-")[0])
	endpos=int(text.split("-")[1])
	positiondict.setdefault(seqid,[startpos,endpos])
ITSxpositionfile.close()	
seqrecords=list(SeqIO.parse(fastafilename, "fasta"))
newseqrecords=[]
for seqrecord in seqrecords:
	newseqrecord=seqrecord
	seq=seqrecord.seq
	if not seqrecord.id in positiondict.keys():
		continue
	position=positiondict[seqid]
	newseq=seq[position[0]-1:position[1]]
	newseqrecord.seq=newseq
	newseqrecords.append(newseqrecord)
#write
SeqIO.write(newseqrecords,output,"fasta")	






#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mon Apr 19 11:56:08 2021

@author: duong
"""
import sys
if sys.version_info[0] >= 3:
	unicode = str
import json
import os, argparse
from Bio import SeqIO

parser=argparse.ArgumentParser(prog='compareCutoffs.py',  
							   usage="%(prog)s [options] -i cutofffiles -o output",
							   description='''Script that compares the cutoffs from different cutoff files.''',
							   epilog="""Written by Duong Vu duong.t.vu@gmail.com""",
   )

parser.add_argument('-i','--filenames', required=True, help='the cutoff filenames separated by a comma')
parser.add_argument('-o','--out', required=True, help='The output.')

args=parser.parse_args()
filenames= args.filenames.split(",")
output=args.out

def LoadCutoffs(cutoffdict,cutofffilename):
	cutoffs={}
	if cutoffsfilename!="" and cutoffsfilename!=None:
		with open(cutoffsfilename) as cutoffsfile:
			cutoffs = json.load(cutoffsfile)
		for rank in cutoffs.keys():
			if not (rank in cutoffdict.keys()):
				cutoffdict.setdefault(rank,{})
			combineddatasets=cutoffdict[rank]
			datasets=cutoffs[rank]
			for datasetname in datasets.keys():
				dataset=datasets[datasetname]
				cutoff=dataset["cut-off"]
				confidence=dataset["confidence"]
				if not (datasetname in combineddatasets.keys()):
					combineddatasets.setdefault(datasetname,{})
				tmp=combineddatasets[datasetname]
				if not (cutofffilename in tmp.keys()):
					tmp.setdefault(cutofffilename,{})	
				tmpdataset=tmp[cutofffilename]	
				tmpdataset["cut-off"]=cutoff	
				tmpdataset["confidence"]=confidence

########MAIN###############

header="Rank\tTaxon name"
cutoffdict={}
for cutoffsfilename in filenames:
	LoadCutoffs(cutoffdict,cutoffsfilename)	
	header=header + "\t" + os.path.basename(cutoffsfilename) + " cut-off\t" + os.path.basename(cutoffsfilename) + " confidence"
header=header + "\n"
outputfile=open(output,"w")
outputfile.write(header)
confidencecounts=[0]*len(filenames)
cutoffcounts=[0]*len(filenames)
texts={}
cutofftexts={}
count=0

for rank in cutoffdict.keys():
	combineddatasets=cutoffdict[rank]
	for datasetname in combineddatasets.keys():
		tmp=combineddatasets[datasetname]
		line=rank + "\t" + datasetname
		maxconfidence=0
		maxcutoff=0
		bestcutoff=0
		bestregion=""
		maxregion=""
		check=0
		text=rank + "\t" + datasetname + "\t"
		confidencelist=[]
		cutofflist=[]
		for filename in filenames:
			cutoff=""
			confidence=""
			if filename in tmp.keys():
				check=check+1
				dataset=tmp[filename]
				cutoff=str(dataset["cut-off"])
				confidence=str(dataset["confidence"])
				confidencelist.append(float(confidence))
				cutofflist.append(float(cutoff))
				if float(confidence) > maxconfidence:
					maxconfidence=float(confidence)
					bestregion=filename
					bestcutoff=float(cutoff)
				if float(cutoff) > maxcutoff:
					maxcutoff=float(cutoff)
					maxregion=filename
			line=line+"\t" + cutoff+"\t"+confidence
		line=line+"\n"	
		outputfile.write(line)	
		if check==len(filenames) :
			#absolute confidence
			#text=text + bestregion  + "\t" + str(bestcutoff) + "\t" + str(maxconfidence)
			if confidencelist.count(maxconfidence)== 1:
				if not (bestregion in texts.keys()):
					texts.setdefault(bestregion,[line])
				else:
					texts[bestregion].append(line)
				confidencecounts[filenames.index(bestregion)]=confidencecounts[filenames.index(bestregion)] + 1
			if cutofflist.count(maxcutoff)== 1:
				if not (maxregion in cutofftexts.keys()):
					cutofftexts.setdefault(maxregion,[line])
				else:
					cutofftexts[maxregion].append(line)
				cutoffcounts[filenames.index(maxregion)]=cutoffcounts[filenames.index(maxregion)] + 1 	 
			count=count+1
		
outputfile.close()		
print("The comparison is saved in file " + output + ".")
print("Number of taxa have all regions available: " + str(count))
i=0
for filename in filenames:
	print("Number of taxa that " + filename + " has the highest resolving power: " + str(confidencecounts[i]) + " (" + str(round(confidencecounts[i]*100/count,2)) + "% )")
	if filename in texts.keys():
		suboutputname=output + "." + os.path.basename(filename) + ".higherconfidence"
		suboutputfile=open(suboutputname,"w")
		#suboutputfile.write("Rank\tTaxon name\tcut-off\tconfidence\n")
		suboutputfile.write(header)
		for line in texts[filename]:
			suboutputfile.write(line)
		suboutputfile.close()	
		print("The taxa that " + filename + " has the highest resolving power are saved in file " + suboutputname)
	i=i+1
i=0
for filename in filenames:
	print("Number of taxa that " + filename + " has the highest similarity cut-offs: " + str(cutoffcounts[i]) + " (" + str(round(cutoffcounts[i]*100/count,2)) + "% )")
	if filename in cutofftexts.keys():
		suboutputname=output + "." + os.path.basename(filename) + ".highercutoff"
		suboutputfile=open(suboutputname,"w")
		#suboutputfile.write("Rank\tTaxon name\tcut-off\tconfidence\n")
		suboutputfile.write(header)
		for line in cutofftexts[filename]:
			suboutputfile.write(line)
		suboutputfile.close()	
		print("The taxa that " + filename + " has the highest cutoff are saved in file " + suboutputname)
	i=i+1	
outputfile.close()
	
		


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Thu Jul  1 15:13:18 2021

@author: duong vu duong.t.vu@gmail.com/d.vu@wi.knaw.nl
"""
import sys, os, subprocess, inspect
path = os.path.dirname(os.path.abspath(inspect.getfile(inspect.currentframe())))
#import lib.library as lib

#git_version = lib.git_version()
#base_version = '1.0.0'
#if git_version:
#    version = base_version+'-'+git_version
#else:
#    version = base_version

default_help = """

Usage:       dnabarcoder <command> <arguments>
version:     %s

Description: dnabarcoder is a tool for the analysis, visualization, classification, and predictions of similarity cut-offs for dna barcodes
    
Command:     overview     	                 Get an overview of the dataset 
             length                          Compute length distribution
             distribution                    Compute sequence distribution
             variation                       Compute sequence variation
             sim                             Compute similarity matrix	                          
             visualize                       Visualize the sequences
             tree                            Create a phylogenetic tree of the sequences			 
             cluster                         Cluster the sequences
             predict                         Predict similarity cut-offs for the sequences
             best                            Compute best similarity cut-offs for the sequences			 
             remove                          Remove similar sequences of the same complexes based on a give threshold
             search                          Search for best matches of the sequences against a file of reference sequences
             classify                        Classify the sequences to the group of their best match if the score is greater than the given cutoff
             verify                          Verify the assigned sequences based on the phylogenetic tree branch lengths			 
             krona                           Visualize classification results using Krona
             evaluate                        Compute accuracy for classification results
			       
Written by Duong Vu duong.t.vu@gmail.com/d.vu@wi.knaw.nl
        """ #% version
wrongcommand=False		
if len(sys.argv) > 1:
	if sys.argv[1] == 'overview':
		help = """
Usage:       dnabarcoder %s <arguments>
version:     %s

Description: The script gives an overview about the barcode dataset
    
Arguments:   -i, --input             Fasta file of Dna sequences, required
             -c, --classification   The taxonomic classification file in tab delimited format, required
             -o, --out               The output folder, default= "dnabarcoder"
Written by Duong Vu duong.t.vu@gmail.com/d.vu@wi.knaw.nl
		""" # % (sys.argv[1], version)
		arguments = sys.argv[2:]
		if len(arguments) > 1:
			cmd = os.path.join(path, 'analysis', 'overview.py')
			arguments.insert(0, cmd)
			exe = sys.executable
			arguments.insert(0, exe)
			subprocess.call(arguments)
		else:
			print(help)
			sys.exit(1)	
	elif sys.argv[1] == 'length':
		help = """
Usage:       dnabarcoder %s <arguments>
version:     %s

Description: The script computes the barcode length distribution
    
Arguments:   -i, --input             Fasta file of Dna sequences, required
             -l, --intervallength    The length of intervals, default=100
             -o, --out               The output folder, default= "dnabarcoder"
Written by Duong Vu duong.t.vu@gmail.com/d.vu@wi.knaw.nl
		""" # % (sys.argv[1], version)
		arguments = sys.argv[2:]
		if len(arguments) > 1:
			cmd = os.path.join(path, 'analysis', 'computeLengthDistribution.py')
			arguments.insert(0, cmd)
			exe = sys.executable
			arguments.insert(0, exe)
			subprocess.call(arguments)
		else:
			print(help)
			sys.exit(1)	
	elif sys.argv[1] == 'distribution':
		help = """
Usage:       dnabarcoder %s <arguments>
version:     %s

Description: The script computes the distribution of the sequences
    
Arguments:   -i, --input             		Fasta file of Dna sequences, required
             -c, --classification    		The taxonomic classification file in tab delimited format, required
             -ranks, --classificationranks 	The classification ranks for computing sequence distribution, required			 
             -n, --numberofdisplayedlabels  The number of the largest taxa to be displayed in the figure, default=8	
             -method, --visualizationmethod The visualization method: krona or plot, default=plot		 			 
             -o, --out                      The output folder, default= "dnabarcoder"
Written by Duong Vu duong.t.vu@gmail.com/d.vu@wi.knaw.nl
		""" # % (sys.argv[1], version)
		arguments = sys.argv[2:]
		if len(arguments) > 1:
			cmd = os.path.join(path, 'analysis', 'computeDistribution.py')
			arguments.insert(0, cmd)
			exe = sys.executable
			arguments.insert(0, exe)
			subprocess.call(arguments)
		else:
			print(help)
			sys.exit(1)		
	elif sys.argv[1] == 'variation':
		help = """
Usage:       dnabarcoder %s <arguments>
version:     %s

Description: The script computes the variation of the sequences
    
Arguments:   -i, --input             	    FASTA file of Dna sequences, required
             -c, --classification    	    The taxonomic classification file in tab delimited format, required
             -ranks, --classificationranks  The classification ranks for computing sequence variation, default=""		 
             -ml, --minalignmentlength      Minimum sequence alignment length required for BLAST, default=400. For short barcode sequences like ITS2 (ITS1) sequences, ml should be set to smaller, 50 for instance.	
             -m, --maxSeqNo                 The maximum number of randomly selected sequences to be computed in the case the groups are too big, default=0 (no maximum is given).		 	
             -plt, --plottype               The type of visualization: boxplot or plot, default=boxplot
             -o, --out                      The output folder, default= "dnabarcoder"			 
Written by Duong Vu duong.t.vu@gmail.com/d.vu@wi.knaw.nl
		""" # % (sys.argv[1], version)
		arguments = sys.argv[2:]
		if len(arguments) > 1:
			cmd = os.path.join(path, 'analysis', 'computeVariation.py')
			arguments.insert(0, cmd)
			exe = sys.executable
			arguments.insert(0, exe)
			subprocess.call(arguments)
		else:
			print(help)
			sys.exit(1)
	elif sys.argv[1] == 'sim':
		help = """
Usage:       dnabarcoder %s <arguments>
version:     %s

Description: The script computes BLAST similarity matrix for sequences
    
Arguments:   -i, --input             	   Fasta file of Dna sequences, required
             -ml, --minalignmentlength      Minimum sequence alignment length required for BLAST, default=400. For short barcode sequences like ITS2 (ITS1) sequences, ml should be set to smaller, 50 for instance.	
             -ms, --minsim                 The minimum similarity that will be saved.
             -o, --out                     The output folder, default= "dnabarcoder"			 
Written by Duong Vu duong.t.vu@gmail.com/d.vu@wi.knaw.nl
		""" # % (sys.argv[1], version)
		arguments = sys.argv[2:]
		if len(arguments) > 1:
			cmd = os.path.join(path, 'analysis', 'computeSim.py')
			arguments.insert(0, cmd)
			exe = sys.executable
			arguments.insert(0, exe)
			subprocess.call(arguments)
		else:
			print(help)
			sys.exit(1)	
	elif sys.argv[1] == 'visualize':
		help = """
Usage:       dnabarcoder %s <arguments>
version:     %s

Description: The script visualizes the sequences based on BLAST similarity 
    
Arguments:   -i, --input             	   Fasta file of Dna sequences, required
             -c, --classification    	   The taxonomic classification file in tab delimited format, required
             -rank, --classificationrank    The classification rank for coloring the sequences
             -coord, --coordinates          The coordinates file in json format if exists.
             -sim, --simfilename            The similarity file if exists.				 
             -ml, --minalignmentlength      Minimum sequence alignment length required for BLAST, default=400. For short barcode sequences like ITS2 (ITS1) sequences, ml should be set to smaller, 50 for instance.	
             -ms, --minsim                  The minimum similarity that will be saved for the similarity matrix, default=0.5			 
             -dim, --dimension              The dimension 2D or 3D for visualization, default=3
             -kneigh, --kneigbors           The k-neighbors number for visualization, default=150
             -size, --size                  The size of the dot
			 -method, --visualizationmethod The visualization method: DiVE and plot, default=plot			
             -o, --out                      The output folder, default= "dnabarcoder"			 
Written by Duong Vu duong.t.vu@gmail.com/d.vu@wi.knaw.nl
		""" # % (sys.argv[1], version)
		arguments = sys.argv[2:]
		if len(arguments) > 1:
			cmd = os.path.join(path, 'visualization', 'visualize.py')
			arguments.insert(0, cmd)
			exe = sys.executable
			arguments.insert(0, exe)
			subprocess.call(arguments)
		else:
			print(help)
			sys.exit(1)	
	elif sys.argv[1] == 'tree':
		help = """
Usage:       dnabarcoder %s <arguments>
version:     %s

Description: The script visualizes the sequences based on BLAST similarity 
    
Arguments:   -i, --input             	   	     FASTA file of Dna sequences, required
             -c, --classification    	         The taxonomic classification file in tab delimited format 
             -ranks, --classificationranks       The classification ranks for getting sequence descriptions
             -o, --out                           The output folder, default= "dnabarcoder"			 
Written by Duong Vu duong.t.vu@gmail.com/d.vu@wi.knaw.nl
		""" # % (sys.argv[1], version)
		arguments = sys.argv[2:]
		if len(arguments) > 1:
			cmd = os.path.join(path, 'visualization', 'maketree.py')
			arguments.insert(0, cmd)
			exe = sys.executable
			arguments.insert(0, exe)
			subprocess.call(arguments)
		else:
			print(help)
			sys.exit(1)				
	elif sys.argv[1] == 'cluster':
		help = """
Usage:       dnabarcoder %s <arguments>
version:     %s

Description: The script cluster the sequences with a given threshold (cutoff)
    
Arguments:   -i, --input             	        FASTA file of Dna sequences, required
             -t, --cutoff            	        The threshold (cutoff) used for clustering, default=0.97            
             -sim, --simfilename                The similarity file if exists				 
             -ml, --minalignmentlength          Minimum sequence alignment length required for BLAST in case simmatrix is not given, default=400. For short barcode sequences like ITS2 (ITS1) sequences, ml should probably be set to smaller, 50 for instance.	
             -c, --classification    	        The taxonomic classification file in tab delimited format, optional. This is used to compute the confidence of clustering.
             -rank, --classificationrank        The classification rank to compute the confidence measure for clustering, optional
             -o, --out                          The output folder, default= "dnabarcoder"			 
Written by Duong Vu duong.t.vu@gmail.com/d.vu@wi.knaw.nl
		""" # % (sys.argv[1], version)
		arguments = sys.argv[2:]
		if len(arguments) > 1:
			cmd = os.path.join(path, 'prediction', 'cluster.py')
			arguments.insert(0, cmd)
			exe = sys.executable
			arguments.insert(0, exe)
			subprocess.call(arguments)
		else:
			print(help)
			sys.exit(1)		
	elif sys.argv[1] == 'remove':
		help = """
Usage:       dnabarcoder %s <arguments>
version:     %s

Description: The script removes sequences of the same complexes based on a given threshold (cutoff)
    
Arguments:   -i, --input             	        FASTA file of Dna sequences, required
             -t, --cutoff            	        The threshold (cutoff) used for clustering, default=0.97            
             -sim, --simfilename                The similarity matrix file if exists				 
             -ml, --minalignmentlength          Minimum sequence alignment length required for BLAST in case simmatrix is not given, default=400. For short barcode sequences like ITS2 (ITS1) sequences, ml should probably be set to smaller, 50 for instance.	
             -c, --classification    	        The taxonomic classification file in tab delimited format, optional. This is used to compute the confidence of clustering
             -rank, --classificationrank        The classification rank to remove complexes.
             -o, --out                          The output folder, default= "dnabarcoder"			 
Written by Duong Vu duong.t.vu@gmail.com/d.vu@wi.knaw.nl
		""" # % (sys.argv[1], version)
		arguments = sys.argv[2:]
		if len(arguments) > 1:
			cmd = os.path.join(path, 'prediction', 'removeComplexes.py')
			arguments.insert(0, cmd)
			exe = sys.executable
			arguments.insert(0, exe)
			subprocess.call(arguments)
		else:
			print(help)
			sys.exit(1)	
	elif sys.argv[1] == 'predict':
		help = """
Usage:       dnabarcoder %s <arguments>
version:     %s

Description: The script predicts similarity cut-offs for sequence identification
    
Arguments:   -i, --input             	                   Fasta file of Dna sequences, required       
             -sim, --simfilename                           The similarity matrix file if exists				 
             -ml, --minalignmentlength                     Minimum sequence alignment length required for BLAST in case simmatrix is not given, default=400. For short barcode sequences like ITS2 (ITS1) sequences, ml should probably be set to smaller, 50 for instance.	
             -c, --classification    	                   The taxonomic classification file in tab delimited format 			 
             -ranks, --classificationranks                 The classification ranks for prediction, separated by ","
             -higherranks, --higherclassificationranks     The higher classification ranks for the prediction. If hp=="", the whole dataset will be used for prediction
             -st, --startingthreshold                      The starting threshold of the prediction
             -et, --endthreshold                           The ending threshold of the prediction		
             -s, --step       	           ,               The step for the prediction				  
             -minGroupNo,--minimumgroupnumber              The minimum number of groups for prediction, default=5
             -minSeqNo,--minimumsequencenumber             The minimum number of sequences for prediction, default=50	
             -redo,--redo                                  Redo the prediction if redo !="", default=""		
             -prefix,--prefix                              Prefix of all output files, default as the base of the input file				  
             -o, --out                                     The output folder, default= "dnabarcoder"			 
Written by Duong Vu duong.t.vu@gmail.com/d.vu@wi.knaw.nl

		""" # % (sys.argv[1], version)
		arguments = sys.argv[2:]
		if len(arguments) > 1:
			cmd = os.path.join(path, 'prediction', 'predict.py')
			arguments.insert(0, cmd)
			exe = sys.executable
			arguments.insert(0, exe)
			subprocess.call(arguments)
		else:
			print(help)
			sys.exit(1)		
	elif sys.argv[1] == 'best':
		help = """
Usage:       dnabarcoder %s <arguments>
version:     %s

Description: The script predicts similarity cut-offs for sequence identification
    
Arguments:   -i, --input             	        the cutoffs file, required       
             -c, --classification    	        The taxonomic classification file in tab delimited format 			 		            
			 -prefix,--prefix                   Prefix of all output files, default as the base of the input file				  
             -o, --out                          The output folder, default= "dnabarcoder"			 
Written by Duong Vu duong.t.vu@gmail.com/d.vu@wi.knaw.nl

		""" # % (sys.argv[1], version)
		arguments = sys.argv[2:]
		if len(arguments) > 1:
			cmd = os.path.join(path, 'prediction', 'computeBestCutoffs.py')
			arguments.insert(0, cmd)
			exe = sys.executable
			arguments.insert(0, exe)
			subprocess.call(arguments)
		else:
			print(help)
			sys.exit(1)				
	elif sys.argv[1] == 'search':
		help = """
Usage:       dnabarcoder %s <arguments>
version:     %s

Description: The script classifies a fasta file of sequences against a reference file of sequences
    
Arguments:   -i, --input             	        Fasta file of Dna sequences to be classified, required       
             -r, --reference                    Fasta file of reference sequences, required    			 
             -ml, --minalignmentlength          Minimum sequence alignment length required for BLAST in case simmatrix is not given, default=400. For short barcode sequences like ITS2 (ITS1) sequences, ml should probably be set to smaller, 50 for instance.	
             -prefix,--prefix                   Prefix of all output files, default as the base of the input file				  
             -o, --out                          The output folder, default= "dnabarcoder"			 
Written by Duong Vu duong.t.vu@gmail.com/d.vu@wi.knaw.nl
		""" # % (sys.argv[1], version)
		arguments = sys.argv[2:]
		if len(arguments) > 1:
			cmd = os.path.join(path, 'classification', 'search.py')
			arguments.insert(0, cmd)
			exe = sys.executable
			arguments.insert(0, exe)
			subprocess.call(arguments)
		else:
			print(help)
			sys.exit(1)				
	elif sys.argv[1] == 'classify':
		help = """
Usage:       dnabarcoder %s <arguments>
version:     %s

Description: The script classifies the sequences to their best match based on local cutoffs or a given global cutoff. For a sequence, a BLAST similarity score to the sequences of the predicted group is computed. The sequence is assigned if the BLAST similarity score is greater or equal than the similarity cut-off
    
Arguments:   -i, --input             	        The file of classified sequences, required
             -f, --fasta                        The fasta file of classified sequences, required    
             -r, --reference                    The fasta file of reference sequences, required    			 
             -ml, --minalignmentlength          Minimum sequence alignment length required for BLAST in case simmatrix is not given, default=400. For short barcode sequences like ITS2 (ITS1) sequences, ml should probably be set to smaller, 50 for instance.	
             -mp, --minproba                    Only consider the classified sequences with a probability greater or equal minproba
             -m, --maxseqno                     Maximum number of the sequences of the predicted taxon name from the classification file will be selected for the comparison to find the best match. If it is not given, all the sequences will be selected, default=0
             -c, --classification    	        The taxonomic classification file in tab delimited format 			 
             -cutoffs, --cutoffs                The similarity cutoffs file predicted by dnabarcoder predict if exists
             -cutoff, --cutoff                  The similarity cutoff, default=0, only used if the similarity cutoffs file is not given
             -confidence, --confidence          The confidence of the similarity cutoff if exists
             -rank, --rank                      The rank to classify the sequences, default=""
             -prefix,--prefix                   Prefix of all output files, default as the base of the input file				  
             -minGroupNo,--minimumgroupnumber   The minimum number of groups for prediction, default=5
             -minSeqNo,--minimumsequencenumber  The minimum number of sequences for prediction, default=50	
             -o, --out                          The output folder, default= "dnabarcoder"			 
Written by Duong Vu duong.t.vu@gmail.com/d.vu@wi.knaw.nl
		""" # % (sys.argv[1], version)
	
		arguments = sys.argv[2:]
		if len(arguments) > 1:
			cmd = os.path.join(path, 'classification', 'classify.py')
			arguments.insert(0, cmd)
			exe = sys.executable
			arguments.insert(0, exe)
			subprocess.call(arguments)
		else:
			print(help)
			sys.exit(1)		
	elif sys.argv[1] == 'verify':
		help = """
Usage:       dnabarcoder %s <arguments>
version:     %s

Description: The script assigns classified sequences based on local cutoffs or a given global cutoff. For a sequence, a BLAST similarity score to the sequences of the predicted group is computed. The sequence is assigned if the BLAST similarity score is greater or equal than the similarity cut-off
    
Arguments:   -i, --input             	        The file of classified sequences, required
             -f, --fasta                        The fasta file of classified sequences, required    
             -r, --reference                    The fasta file of reference sequences, required    			 
             -m, --maxseqno                     Maximum number of the sequences of the predicted taxon name from the classification file will be selected for the comparison to find the best match. If it is not given, all the sequences will be selected, default=0
             -c, --classification    	        The taxonomic classification file in tab delimited format 			 
             -rank, --rank                      The rank to classify the sequences, default=""
             -prefix,--prefix                   Prefix of all output files, default as the base of the input file				  
             -o, --out                          The output folder, default= "dnabarcoder"			 
Written by Duong Vu duong.t.vu@gmail.com/d.vu@wi.knaw.nl
		""" # % (sys.argv[1], version)
	
		arguments = sys.argv[2:]
		if len(arguments) > 1:
			cmd = os.path.join(path, 'classification', 'verify.py')
			arguments.insert(0, cmd)
			exe = sys.executable
			arguments.insert(0, exe)
			subprocess.call(arguments)
		else:
			print(help)
			sys.exit(1)				
	elif sys.argv[1] == 'krona':
		help = """
Usage:       dnabarcoder %s <arguments>
version:     %s

Description: The script visualizes the classification/assignment results using Krona
    
Arguments:   -i, --input             	        The file of classified/assigned sequences, required
             -c, --classification               The fasta file of classified sequences, required 
             -o, --out                          The output folder, default= "dnabarcoder"			 
Written by Duong Vu duong.t.vu@gmail.com/d.vu@wi.knaw.nl
		""" # % (sys.argv[1], version)
		arguments = sys.argv[2:]
		if len(arguments) > 1:
			cmd = os.path.join(path, 'classification', 'visualizeClassification.py')
			arguments.insert(0, cmd)
			exe = sys.executable
			arguments.insert(0, exe)
			subprocess.call(arguments)
		else:
			print(help)
			sys.exit(1)	
	elif sys.argv[1] == 'evaluate':
		help = """
Usage:       dnabarcoder %s <arguments>
version:     %s

Description: The script compute classification metrices
    
Arguments:   -i, --input             	        The file of classified/assigned sequences, required
             -c, --queryclassification          The classification file of query sequences, required
             -r, --refclassification            The classification file of reference sequences, required
             -o, --out                          The output folder, default= "dnabarcoder"			 
Written by Duong Vu duong.t.vu@gmail.com/d.vu@wi.knaw.nl
		""" # % (sys.argv[1], version)		
		arguments = sys.argv[2:]
		if len(arguments) > 1:
			cmd = os.path.join(path, 'classification', 'evaluate.py')
			arguments.insert(0, cmd)
			exe = sys.executable
			arguments.insert(0, exe)
			subprocess.call(arguments)
		else:
			print(help)
			sys.exit(1)				
	else:
		wrongcommand=True
if len(sys.argv) == 1 or wrongcommand==True:
	print(default_help)
	


#!/usr/bin/env python
# FILE: computeDistribution.py
# AUTHOR: Duong Vu
# CREATE DATE: 07 June 2019
import sys, argparse
if sys.version_info[0] >= 3:
	unicode = str
import numpy as np
import os
from Bio import SeqIO
import json
import random
import matplotlib.pyplot as plt
plt.rc('font',size=6)

import multiprocessing
nproc=multiprocessing.cpu_count()

parser=argparse.ArgumentParser(prog='computeDistribution.py',  
							   usage="%(prog)s [options] -i fastafile -c classificationfilename -ranks classificationranks  -o output",
							   description='''Script that computes the distribution of the sequences based on given classification. ''',
							   epilog="""Written by Duong Vu duong.t.vu@gmail.com""",
   )

parser.add_argument('-i','--input', required=True, help='the fasta file to be clustered.')
parser.add_argument('-o','--out', default="dnabarcoder", help='The output folder.')
parser.add_argument('-c','--classification', help='the classification file in tab. format.')
parser.add_argument('-ranks','--classificationranks', default="", help='the classification ranks to compute distribution, separated by ",".')
parser.add_argument('-n','--numberofdisplayedlabels', type=int, default=5, help='The number of labels to be displayed.')
parser.add_argument('-labelstyle','--labelstyle', default='normal', help='The label style to be displayed: normal, italic, or bold.')
parser.add_argument('-method','--visualizationmethod', default="plot", help='The visualization method. There are two methods to be selected: krona and plot.')
parser.add_argument('-prefix','--prefix',default="", help='The prefix of the output files.')

args=parser.parse_args()
referencename= args.input
classificationfilename=args.classification
jsonvariationfilename =args.out
labelno=args.numberofdisplayedlabels
method=args.visualizationmethod
outputpath=args.out
prefix=args.prefix

if not os.path.exists(outputpath):
	os.system("mkdir " + outputpath)

def GetBase(filename):
	if not ("." in filename):
		return filename
	return filename[:-(len(filename)-filename.rindex("."))] 

def GetWorkingBase(filename):
	basename=os.path.basename(filename)
	if "." in basename:
		basename=basename[:-(len(basename)-basename.rindex("."))] 
	path=outputpath + "/" + basename
	return path

def LoadClassification(seqIDs,seqrecords,classificationfilename,poslist):
	classification={}
	if classificationfilename == "":
		classification["unidentified"]=len(seqrecords)
		return classification
	records= list(open(classificationfilename, "r"))
	feature=""
	for line in records:
		if line.startswith("#"):
			i=0 
			for pos in poslist:
				if i==0:
					feature=line.rstrip().split("\t")[pos]
				else:
					feature=feature + line.rstrip().split("\t")[pos]
				i=i+1
			continue 
		elements=line.split("\t")
		seqid = elements[0].replace(">","").rstrip()
		if seqid in seqIDs:
			classname=""
			i=0
			for pos in poslist:
				taxonname=elements[pos].rstrip()
				if taxonname=="":
					taxonname="unidentified"
				if i==0:
					classname=taxonname
				else:
					classname=classname + "\t" + taxonname
				i=i+1	
			if classname in classification.keys():
				classification[classname]=classification[classname] + 1
			else:
				classification.setdefault(classname, 1)
	return classification,feature

def SaveDistributionInTabFormat(output,classification):
	outputfile=open(output,"w")
	outputfile.write("Taxonname\tNumber of sequences\t%\n")	
	values = classification.values() 
	total= sum(values)
	sorted_classification = sorted(classification.items(), key=lambda x: x[1], reverse=True)
	i=0
	count=0
	for item in sorted_classification:
		classname=item[0]
		seqno=item[1]
		percentage=round(seqno*100/total,2)
		outputfile.write(classname + "\t" + str(seqno) + "\t" + str(percentage) + "\n")
		if i<args.numberofdisplayedlabels:
			count=count+seqno
		i=i+1
	percentage=round(count*100/total,2)
	print("The first " + str(args.numberofdisplayedlabels) + " groups of " + output + " consist of " + str(count) + "/" + str(total) + " (" +  str(percentage) + "%) sequences")	
	outputfile.close()
	
def PlotPieChart(figoutput,title,classification,displayed):
	values = classification.values() 
	total= sum(values)
	unseqno=0
	if "unidentified" in classification.keys():
		unseqno=classification["unidentified"]
		del classification['unidentified']
	#sort variations based on median thresholds with decreasing order
	sorted_classification = sorted(classification.items(), key=lambda x: x[1], reverse=True)
	classnames=[]
	seqnos=[]
	for item in sorted_classification:
		seqno=item[1]
		p=round(seqno*100/total,2)
		classname=item[0] + " " + str(p) + "%" 
		classnames.append(classname)
		seqnos.append(seqno)
	if unseqno > 0:
		p=round(unseqno*100/total,2)
		classnames.append("unidentified" + " " + str(p) + "%" )	
		seqnos.append(unseqno)
	#select labels to show
	n=len(classnames)
	if labelno > 0:
		n=min(labelno,len(classnames))
	displayedlabels=classnames[0:n]
	cmap = plt.get_cmap("tab20c")
	colors = plt.cm.Set1(np.linspace(0, 1, labelno)) #prism
	if labelno < len(classnames):
		colors1=cmap(np.arange(len(classnames)-labelno)) 
		#colors1 = plt.cm.rainbow(np.linspace(0, 1, (len(classnames)-labelno)))
		colors=np.concatenate((colors,colors1), axis=0)
	if unseqno >0:	
		colors[len(classnames)-1]= np.array([0,0,0,1])#black	
	#pie chart
	fig, ax = plt.subplots(figsize=(3,3))
	#ax.pie(seqnos, startangle = 90, wedgeprops = { 'linewidth': 2, "edgecolor" :"k" }) #autopct='%.2f'
	ax.set_title
	ax.pie(seqnos,colors=colors,startangle = 90) #autopct='%.2f'
	#plt.title(title , horizontalalignment='center')
	plt.gca().axis("equal")
	plt.legend(displayedlabels,  loc="lower right", bbox_transform=plt.gcf().transFigure,prop={'style': args.labelstyle})#bbox_to_anchor=(1,1)
	plt.rcParams['font.size'] = 6.0
	plt.tight_layout()
	#plt.subplots_adjust(left=0.1, bottom=0.1, right=0.5)
	plt.savefig(figoutput, dpi = 500)
	if displayed==True:
		plt.show()

def PlotNestedPieCharts(figoutput,title,classificationlist,labels):
	#colors = plt.cm.Set1(np.linspace(0, 1,len(data)))	
	# create a figure with two subplots
	#pie chart
	fig, ax = plt.subplots(figsize=(3,3))
	ax.set_title(title)
	size = 0.1
	r=1
	i=0
	sorted_classificationlist=sorted(classificationlist, key=lambda x: len(x))
	##print(sorted_classificationlist)
	displayedlabels=[]
	cmap = plt.get_cmap("tab20c")
	for classification in sorted_classificationlist:
		values = classification.values() 
		total= sum(values)
		unseqno=0
		if "unidentified" in classification.keys():
			unseqno=classification["unidentified"]
			del classification['unidentified']
		sorted_classification = sorted(classification.items(), key=lambda x: x[1], reverse=True)
		classnames=[]
		seqnos=[]
		for item in sorted_classification:
			seqno=item[1]	
			p=round(seqno*100/total,2)
			classname=item[0] + " " + str(p) + "%" 
			classnames.append(classname)
			seqnos.append(seqno)
		if unseqno > 0:
			p=round(unseqno*100/total,2)
			classnames.append("unidentified" + " " + str(p) + "%" )	
			seqnos.append(unseqno)	
		#pie chart
		#select labels to show
		if i==0:
		#select labels to show
			n=len(classnames)
			if labelno > 0:
				n=min(labelno,len(classnames))
				displayedlabels=classnames[0:n]
		#plt.pie(seqnos, labels = displayedlabels, startangle = 90)
		#colors=cmap(np.arange(len(classnames))) 
		colors = plt.cm.Set1(np.linspace(0, 1, labelno)) #prism
		if labelno < len(classnames):
			colors1=cmap(np.arange(len(classnames)-labelno)) 
			#colors1 = plt.cm.rainbow(np.linspace(0, 1, (len(classnames)-labelno)))
			colors=np.concatenate((colors,colors1), axis=0)
		if unseqno >0:	
			colors[len(classnames)-1]= np.array([0,0,0,1])#black			
		ax.pie(seqnos, radius=r, startangle = 90, colors=colors, wedgeprops=dict(width=size, edgecolor='w'))
		r=r-size
		i=i+1	
	#plt.title(title, horizontalalignment='center')
	plt.gca().axis("equal")
	plt.legend(displayedlabels,  loc="lower right", bbox_transform=plt.gcf().transFigure,prop={'style': args.labelstyle})#bbox_to_anchor=(1,1)
	plt.rcParams['font.size'] = 6.0
	plt.tight_layout()
	plt.savefig(figoutput, dpi = 500)
	plt.show()
	
def KronaPieCharts(classification,kronareport,kronahtml):
	kronareportfile=open(kronareport,"w")
	for classname in classification.keys():
		kronareportfile.write(str(classification[classname]) + "\t" + classname + "\n")
	kronareportfile.close()	
	#create kronahtml
	command="ImportText.pl " + kronareport + " -o " + kronahtml
	print(command)
	os.system(command)
	os.system("firefox " + kronahtml) 
	
def GetPositionList(classificationfilename,ranks):
	ranklist=[]	
	if "," in ranks:
		ranklist=ranks.split(",")
	elif ranks !="":
		ranklist.append(ranks)
	positionlist=[]
	classificationfile=open(classificationfilename)
	header=classificationfile.readline()
	header=header.rstrip()
	classificationfile.close()
	texts=header.split("\t")
	isError=False
	for rank in ranklist:
		if rank in texts:
			pos=texts.index(rank)
			positionlist.append(pos)
		else:
			print("The rank " + rank + " is not given in the classification." )
			isError=True
	return positionlist,ranklist,isError
##############################################################################
# MAIN
##############################################################################
path=sys.argv[0]
path=path[:-(len(path)-path.rindex("/")-1)]
displayed=True
poslist=[]

if prefix=="":
	prefix=GetBase(os.path.basename(referencename))

poslist,ranklist,isError=GetPositionList(classificationfilename,args.classificationranks)	
if isError==True:
	sys.exit()
displayed=False
if len(poslist)==1:
	displayed=True
	
#load train seq records
referencerecords =  list(SeqIO.parse(referencename, "fasta"))
referenceIDs=[]
for seq in referencerecords:
	referenceIDs.append(seq.id)
if method!="krona":
	classificationlist=[]
	labels=[]
	for classificationposition in poslist:
		jsonfilename = GetWorkingBase(prefix) + "." + str(classificationposition) + ".distribution"
		figoutput=GetBase(jsonfilename) + ".distribution.png" 
		#Load classes, classification:
		classification,feature= LoadClassification(referenceIDs,referencerecords,classificationfilename, [classificationposition])
		title=""
		if feature=="":
			title= prefix + ": the distribution of the the sequences of the groups at position " + str(classificationposition)
		elif feature.lower()== "species":
			title=prefix + ": the distribution of the sequences at the species level"		
		elif feature.lower()== "genus":
			title=prefix + ": the distribution of the sequences at the genus level"	
		elif feature.lower()== "family":
			title=prefix + ": the distribution of the sequences at the family level"	
		elif feature.lower()== "order":
			title=prefix + ": the distribution of the sequences at the order level"	
		elif feature.lower()== "class":
			title=prefix + ": the distribution of the sequences at the class level"	
		elif feature.lower()== "phylum":
			title=prefix + ": the distribution of the sequences at the phylum level"	
		elif feature.lower()== "kingdom":
			title=prefix + ": the distribution of the sequences at the kingdom level"	
		#save classification	
		SaveDistributionInTabFormat(jsonfilename + ".txt",classification)	
		newclassification=classification.copy()
		classificationlist.append(newclassification)
		labels.append(feature)
		#plot
		PlotPieChart(figoutput,title,classification,displayed)
		print("The results are saved in the json file  " + jsonfilename + " and tab file " + jsonfilename + ".txt. The figure is saved in " + figoutput + "."  )	
	if len(poslist)>1:
		jsonfilename=""
		jsonfilename = GetWorkingBase(prefix) + ".distribution"
		figoutput=jsonfilename + ".png" 
		print("The figure of the variations of all groups are saved in file " + figoutput + ".")
		title=prefix + ": the distribution of the sequences"
		PlotNestedPieCharts(figoutput,title,classificationlist,labels)
else:
	kronareport = GetWorkingBase(prefix) + ".krona.report"
	kronahtml=GetBase(kronareport) + ".html"
	classification,feature= LoadClassification(referenceIDs,referencerecords,classificationfilename, poslist)
	KronaPieCharts(classification,kronareport,kronahtml)
	print("The krona report and html are saved in iles " + kronareport + " and " + kronahtml + ".") 
		
	
			





#!/usr/bin/env python
# FILE: addtexttosequenceheaders.py, requiring Python 2.7 if classification file contains non-ascii characters
# AUTHOR: Duong Vu
# CREATE DATE: 07 June 2019
import sys, argparse
import numpy as np
import os
from Bio import SeqIO
import json
import multiprocessing
nproc=multiprocessing.cpu_count()
#from keras.utils import np_utils

parser=argparse.ArgumentParser(prog='addTextTosequenceheaders.py',  
							   usage="%(prog)s [options] -i fastafile -t text -o output",
							   description='''Script that adds some default text to sequence headers. ''',
							   epilog="""Written by Duong Vu duong.t.vu@gmail.com""",
   )

parser.add_argument('-i','--input', required=True, help='the fasta file to be clustered.')
parser.add_argument('-o','--out', help='The fasta filename with sequence headers containing the given text.') 
parser.add_argument('-t','--text', required=True, help='the text to be added to the sequence headers')

args=parser.parse_args()
fastafilename= args.input
text=args.text
output=args.out

#fastafilename=sys.argv[1]
#text=sys.argv[2]
#output=sys.argv[3]

outputfile=open(output,"w")
fastafile=open(fastafilename)

for line in fastafile:
	if line.startswith(">"):
		line=line.rstrip()
		header=line + text + "\n"
		outputfile.write(header)	
	else:		
		outputfile.write(line)
fastafile.close()
outputfile.close()		




#!/usr/bin/env python
# FILE: cluster.py
# AUTHOR: Duong Vu
# CREATE DATE: 08 oct 2020
import os
import sys, argparse
from Bio import SeqIO
import multiprocessing

parser=argparse.ArgumentParser(prog='cluster.py',  
							   usage="%(prog)s [options] -i fastafile -t threshold -mc mincoverage -c classificationfilename -p classificationposition -o output",
							   description='''Script that clusters the sequences of the fasta file based on BLAST comparison with the given threshold. The mincoverage is given for BLAST comparision. The classificationfilename and position are given optionally, in order to compute the quality (F-measure) of the clustering. ''',
							   epilog="""Written by Duong Vu duong.t.vu@gmail.com""",
   )

parser.add_argument('-i','--input', required=True, help='the fasta file to be clustered.')
parser.add_argument('-t','--cutoff', type=float, default=0.97, help='The threshold (cutoff) for the classification.')
parser.add_argument('-ml','--minalignmentlength', type=int, default=400, help='Minimum sequence alignment length required for BLAST. For short barcode sequences like ITS2 (ITS1) sequences, minalignmentlength should probably be set to smaller, 50 for instance.')
parser.add_argument('-o','--out', default="dnabarcoder", help='The output folder.')
parser.add_argument('-c','--classification', default="", help='the classification file in tab. format.')
#parser.add_argument('-p','--classificationpos', type=int, default=0, help='the classification position to load the classification.')
parser.add_argument('-rank','--classificationrank', default="", help='the classification rank to evaluate the clustering result.')
parser.add_argument('-sim','--simfilename', help='The similarity matrix of the sequences if exists.')
parser.add_argument('-maxsimmatrixsize','--maxSimMatrixSize', type=int, default=20000, help='The maximum number of sequences to load or compute a full similarity matrix. In case the number of sequences is greater than this number, only similarity values greater than 0 will be loaded to avoid memory problems.')

args=parser.parse_args()
fastafilename= args.input
threshold=args.cutoff
mincoverage = args.minalignmentlength
classificationfilename=args.classification
rank=args.classificationrank
#classificationpos=args.classificationpos
simfilename=args.simfilename
outputpath=args.out
if not os.path.exists(outputpath):
	os.system("mkdir " + outputpath)

nproc=multiprocessing.cpu_count()

def GetBase(filename):
	return filename[:-(len(filename)-filename.rindex("."))] 

def GetWorkingBase(filename):
	basename=os.path.basename(filename)
	basename=basename[:-(len(basename)-basename.rindex("."))] 
	path=outputpath + "/" + basename
	return path

class PointDef:
	def __init__(self, id,flag,neighbors,seqrecord):
		self.id=id
		self.flag=flag
		self.neighbors = neighbors
		self.seqrecord=seqrecord

class ClusterDef:
	def __init__(self, id, pointids):
		self.id=id
		self.pointids=pointids

def GetSeqIndex(seqname,seqlist):
	i=0
	for seq in seqlist:
		if (seqname == seq.id):
			return i
		i = i + 1
	return -1
	
def LoadSim(simfilename):
	simmatrix = {} #we use dictionary to reduce the memory constraints 
	simfile = open(simfilename)
	seqids=[]
	for line in simfile:
		numbers=line.rstrip().split(" ")
		i=numbers[0]
		j=numbers[1]
		seqids.append(i)
		seqids.append(j)
		if i not in simmatrix.keys():
			simmatrix.setdefault(i, {})
		simmatrix[i][j]=float(numbers[2])
	seqids=list(set(seqids))	
	for seqid1 in seqids:
		if not (seqid1 in simmatrix.keys()):
			simmatrix.setdefault(seqid1, {})
			simmatrix[seqid1][seqid1]=1	
		if len(seqids) < args.maxSimMatrixSize: #load full matrix	
			for seqid2 in seqids:
				if not seqid2 in simmatrix[seqid1].keys():
					#simmatrix[seqid1].setdefault(seqid2,0)
					simmatrix[seqid1][seqid2]=0
	simfile.close()		
	return simmatrix
	
def SaveSim(simmatrix,simfilename):
	simfile=open(simfilename,"w")
	for i in simmatrix.keys():
		for j in simmatrix[i].keys():
			simfile.write(str(i) + " " + str(j) + " " + str(simmatrix[i][j]) + "\n")
	simfile.close()	
	
def ComputeSim(fastafilename,seqrecords,mincoverage):
	blastoutput = fastafilename + ".blast.out"		
	blastdb=fastafilename + ".db"		
	#blast
	print("Comparing the sequences of " + fastafilename + " using Blast...")
	makedbcommand = "makeblastdb -in " + fastafilename + " -dbtype \'nucl\' " +  " -out " + blastdb
	print(makedbcommand)
	os.system(makedbcommand)
	blastcommand = "blastn -query " + fastafilename + " -db  " + blastdb + " -task blastn-short -outfmt 6 -out " + blastoutput + " -num_threads " + str(nproc)
	if mincoverage >=400:
		blastcommand = "blastn -query " + fastafilename + " -db " + blastdb + " -outfmt 6 -out " + blastoutput + " -num_threads " + str(nproc)
	print(blastcommand)
	os.system(blastcommand)
	if not os.path.exists(blastoutput):
		print("Cannot compare the sequences of " + fastafilename + " using Blast...")
		logfile=open(GetWorkingBase((os.path.basename(args.input))) + ".predict.log","w")
		logfile.write("Cannot compare the sequences of " + fastafilename + " using Blast...")
		logfile.write("Make BLAST database command: " + makedbcommand + "\n")
		logfile.write("No output for the BLAST command: " + blastcommand + "\n")
		logfile.write("Please rerun prediction for " + os.path.basename(fastafilename) + ".")
		logfile.close()
		return {}
	print("Reading Blast results of " + fastafilename + "...")
	simmatrix={}
	for seqid1 in seqrecords.keys():
		simmatrix.setdefault(seqid1,{})
		for seqid2 in seqrecords.keys():
			if seqid1==seqid2:
				simmatrix[seqid1][seqid2]=1
			elif len(seqrecords.keys()) < args.maxSimMatrixSize: #load full matrix	
				simmatrix[seqid1][seqid2]=0
	#read blast output
	blastoutputfile = open(blastoutput)
	score=0
	for line in blastoutputfile:
		if line.rstrip()=="":
			continue
		words = line.split("\t")
		i = words[0].rstrip()
		j = words[1].rstrip()
		pos1 = int(words[6])
		pos2 = int(words[7])
		iden = float(words[2]) 
		sim=float(iden)/100
		coverage=abs(pos2-pos1)
		score=sim
		if coverage < mincoverage:
			score=float(score * coverage)/mincoverage
		if len(seqrecords.keys()) < args.maxSimMatrixSize: #the full sim matrix has been loaded 
			if simmatrix[i][j] < score:
				simmatrix[i][j]=round(score,4)
				simmatrix[j][i]=round(score,4)
		else:		
			if j in simmatrix[i].keys():
				if simmatrix[i][j] < score:
					simmatrix[i][j]=round(score,4)
					simmatrix[j][i]=round(score,4)
			else:
				simmatrix[i][j]=round(score,4)
				simmatrix[j][i]=round(score,4)	
		#simmatrix[j][i]=score
	os.system("rm " + blastoutput)
	os.system("rm " + blastdb + "*")
	#os.system("rm " + blastdb + ".*")
	return simmatrix

def LoadNeighbors(seqids,subsimmatrix,threshold):
	neighbordict={}
	for seqid in seqids:
		neighbordict.setdefault(seqid, [])
	if len(subsimmatrix.keys()) < args.maxSimMatrixSize:	 #the full matrix has been loaded
		for i in seqids:
			for j in seqids:
				if subsimmatrix[i][j] >= threshold:
					neighbordict[i].append(j)
					neighbordict[j].append(i)
	else:
		for i in seqids:
			for j in seqids:
				if j in subsimmatrix[i].keys():
					if subsimmatrix[i][j] >= threshold:
						neighbordict[i].append(j)
						neighbordict[j].append(i)					
	#os.system("rm out.txt")
	return neighbordict

def LoadPoints(neigbordict,seqrecords):
	points={}
	for seqid in  seqrecords.keys():
		point = PointDef(seqid,False,neigbordict[seqid],seqrecords[seqid])
		points[seqid]=point
	return points

def ExpandCluster(root,cluster,points):
	for i in root.neighbors:
		if points[i].flag==False:
			points[i].flag=True
			cluster.pointids.append(i)
			ExpandCluster(points[i],cluster,points)		

def Cluster(points,clusters):
	for pointid in points.keys():
		if points[pointid].flag==False:
			points[pointid].flag=True
			cluster = ClusterDef(len(clusters),[])
			cluster.pointids.append(pointid)
			ExpandCluster(points[pointid], cluster,points)
			clusters.append(cluster)			
			
def ComputeFmeasure(classes,clusters):
	#compute F-measure
	f=0
	n=0
	for classname in classes.keys():
		group=classes[classname]
		m = 0
		for cluster in clusters:
			i = len(set(group) & set(cluster.pointids))
			v = float(2*i)/float((len(group) + len(cluster.pointids)))
			if m < v:
				m=v		
		n = n + len(group)
		f = f +	(len(group)*m)	
	return float(f)/float(n) 

def LoadClasses(seqids,classificationfilename,pos):
	classificationfile= open(classificationfilename)
#	records= open(classificationfilename,errors='ignore')
	classification={}
	classes={}
	for line in classificationfile:
		if line.startswith("#"):
			 continue
		words=line.split("\t")
		seqid= words[0].rstrip().replace(">","")
		if seqid in seqids:
			classname=""
			if pos < len(words):
				classname=words[pos].rstrip()
			classification[seqid]=classname
			if classname in classes.keys():
				classes[classname].append(seqid)
			else:
				classes.setdefault(classname,[seqid]) 
	classificationfile.close()
	return classes,classification

def SaveClusters(clusters,seqrecords,classification,output):
	outputfile=open(output,"w")
	outputfile.write("ClusterID\tSequenceID\tClassification\tPrediction\n")
	for cluster in clusters:
		classnames=[]
		seqnumbers=[]
		for seqid in cluster.pointids:
			seqrecord = seqrecords[seqid]
			classname=""
			if len(classification) >0:
				classname=classification[seqid]
			if classname in classnames:
				i=classnames.index(classname)
				seqnumbers[i]=seqnumbers[i] +1 
			else:
				classnames.append(classname)
				seqnumbers.append(1)
		classname=""		
		if max(seqnumbers) in seqnumbers:
			maxindex=seqnumbers.index(max(seqnumbers))
			classname=classnames[maxindex]
		for id in cluster.pointids:
			seqrecord = seqrecords[id]
			givenclassname=""
			if len(classification) >0:
				givenclassname=classification[id]
			outputfile.write(str(cluster.id) + "\t" + seqrecord.description + "\t" + givenclassname + "\t" + classname + "\n")		
	outputfile.close()
	
def GetPosition(classificationfilename,rank):
	pos=-1
	classificationfile=open(classificationfilename)
	header=classificationfile.readline()
	header=header.rstrip()
	classificationfile.close()
	texts=header.split("\t")
	isError=False
	if rank in texts:
		pos=texts.index(rank)
	else:
		print("The rank " + rank + " is not given in the classification." )
		isError=True
	return pos,isError

if __name__ == "__main__":
	outputname=GetWorkingBase(fastafilename) + ".clustered"	
	if simfilename=="" or simfilename==None:
		simfilename=GetWorkingBase(fastafilename) + ".sim"
	seqrecords=SeqIO.to_dict(SeqIO.parse(fastafilename, "fasta"))
	sys.setrecursionlimit(len(seqrecords)*2)
	#seqrecords=SeqIO.to_dict(SeqIO.parse(fastafilename, "fasta"))
	classes = []
	classnames = []
	classification=[]
	if classificationfilename!="":
		classificationpos,isError=GetPosition(classificationfilename,rank)
		if isError==False:
			classes,classification=LoadClasses(seqrecords.keys(),classificationfilename,classificationpos)
	#load similarity matrix
	#simmatrix = [[0 for x in range(len(seqrecords))] for y in range(len(seqrecords))]
	simmatrix={} 
	if os.path.exists(simfilename):
		print("Loading similarity matrix " + simfilename)
		simmatrix=LoadSim(simfilename)
	else:	
		print("Computing similarity matrix...")
		simmatrix=ComputeSim(fastafilename,seqrecords,mincoverage)
		print("Save similarity matrix " + simfilename)
		SaveSim(simmatrix,simfilename)
	#load neighbors
	neighbordict = LoadNeighbors(seqrecords.keys(),simmatrix,threshold)
	points=LoadPoints(neighbordict,seqrecords)
	clusters=[]
	print("Clustering...")
	Cluster(points,clusters)
	if len(classes) >0:
		fmeasure=ComputeFmeasure(classes,clusters)
		print("Threshold\tFmeasure")
		print(str(threshold) + "\t" + str(fmeasure))
	print("Saving clusters...")	
	SaveClusters(clusters,seqrecords,classification,outputname)
	print("The clustering result is saved in file " + outputname + ".")





#!/usr/bin/env python
# FILE: removesimilarsequences.py
# AUTHOR: Duong Vu
# CREATE DATE: 02 july 2020
import os
import sys, argparse
from Bio import SeqIO
import multiprocessing
nproc=multiprocessing.cpu_count()
parser=argparse.ArgumentParser(prog='removeComplexes.py',  
							   usage="%(prog)s [options] -i fastafile -t threshold -c classification -p position -out outputname",
							   description='''Script that removes similar sequences of the same complexes based on a given threshold. ''',
							   epilog="""Written by Duong Vu duong.t.vu@gmail.com""",
   )
parser.add_argument('-i','--input', required=True, help='the fasta file to be clustered.')
parser.add_argument('-t','--cutoff', type=float, default=1, help='The threshold for the classification.')
parser.add_argument('-ml','--minalignmentlength', type=int, default=400, help='Minimum sequence alignment length required for BLAST. For short barcode sequences like ITS2 (ITS1) sequences, minalignmentlength should probably be set to smaller, 50 for instance.')
parser.add_argument('-o','--out', default="dnabarcoder", help='The output folder.')
parser.add_argument('-c','--classification', help='the classification file in tab. format.')
#parser.add_argument('-p','--classificationpos', type=int, default=0, help='the classification position to load the classification.')
parser.add_argument('-rank','--classificationrank', default="", help='the classification rank for loading the complexes.')
parser.add_argument('-sim','--simfilename', help='The similarity matrix of the sequences if exists.')


args=parser.parse_args()
fastafilename= args.input
threshold=args.cutoff
mincoverage = args.minalignmentlength
classificationfilename=args.classification
#classificationpos=args.classificationpos
rank=args.classificationrank
outputpath=args.out
simfilename=args.simfilename
if not os.path.exists(outputpath):
	os.system("mkdir " + outputpath)

def GetBase(filename):
	return filename[:-(len(filename)-filename.rindex("."))] 
	
def GetWorkingBase(filename):
	basename=os.path.basename(filename)
	basename=basename[:-(len(basename)-basename.rindex("."))] 
	path=outputpath + "/" + basename
	return path

class PointDef:
	def __init__(self, id,flag,neighbors,seqrecord):
		self.id=id
		self.flag=flag
		self.neighbors = neighbors
		self.seqrecord=seqrecord

class ClusterDef:
	def __init__(self, id, pointids):
		self.id=id
		self.pointids=pointids

def GetSeqIndex(seqname,seqlist):
	i=0
	for seq in seqlist:
		if (seqname == seq.id):
			return i
		i = i + 1
	return -1

def LoadSim(simfilename):
	simmatrix = {} #we use dictionary to reduce the memory constraints 
	simfile = open(simfilename)
	for line in simfile:
		numbers=line.rstrip().split(" ")
		i=numbers[0]
		j=numbers[1]
		if i not in simmatrix.keys():
			simmatrix.setdefault(i, {})
		if j not in simmatrix[i].keys():
			simmatrix[i].setdefault(j, 0)
		if float(numbers[2]) > simmatrix[i][j]:
			simmatrix[i][j]=float(numbers[2])
	simfile.close()		
	return simmatrix

def SaveSim(simmatrix,simfilename):
	simfile=open(simfilename,"w")
	for i in simmatrix.keys():
		for j in simmatrix[i].keys():
			simfile.write(str(i) + " " + str(j) + " " + str(simmatrix[i][j]) + "\n")
	simfile.close()
	
def ComputeSim(fastafilename,seqrecords,mincoverage):
	simmatrix={}
	for seqid in seqrecords.keys():
		simmatrix.setdefault(seqid,{})
		simmatrix[seqid][seqid]=1
	#blast
	makedbcommand = "makeblastdb -in " + fastafilename + " -dbtype \'nucl\' " +  " -out db"
	os.system(makedbcommand)
	blastcommand = "blastn -query " + fastafilename + " -db  db -task blastn-short -outfmt 6 -out out.txt -num_threads " + str(nproc)
	if mincoverage >=300:
		blastcommand = "blastn -query " + fastafilename + " -db  db -outfmt 6 -out out.txt -num_threads " + str(nproc)
	os.system(blastcommand)
	
	#read blast output
	blastoutputfile = open("out.txt")
	score=0
	for line in blastoutputfile:
		if line.rstrip()=="":
			continue
		words = line.split("\t")
		i = words[0].rstrip()
		j = words[1].rstrip()
		pos1 = int(words[6])
		pos2 = int(words[7])
		iden = float(words[2]) 
		sim=float(iden)/100
		coverage=abs(pos2-pos1)
		score=sim
		if coverage < mincoverage:
			score=float(score * coverage)/mincoverage	
		if j in simmatrix[i].keys():
			if simmatrix[i][j] < score:
				simmatrix[i][j]=round(score,4)
				simmatrix[j][i]=round(score,4)
		else:
			simmatrix[i][j]=round(score,4)
			simmatrix[j][i]=round(score,4)
		#simmatrix[j][i]=score
	os.system("rm out.txt")
	return simmatrix

def LoadNeighbors(seqids,simmatrix,threshold):
	neighbordict={}
	for seqid in seqids:
		neighbordict.setdefault(seqid, [])
	for i in seqids:
		for j in seqids:
			if not j in simmatrix[i].keys():
				continue
			if simmatrix[i][j] >= threshold:
				if (j not in neighbordict[i]):
					neighbordict[i].append(j)
				if (i not in neighbordict[j]):
					neighbordict[j].append(i)
	#os.system("rm out.txt")
	return neighbordict

def LoadPoints(neigbordict,seqrecords):
	points={}
	for seqid in  seqrecords.keys():
		point = PointDef(seqid,False,neigbordict[seqid],seqrecords[seqid])
		points[seqid]=point
	return points

def ExpandCluster(root,cluster,points):
	for i in root.neighbors:
		if points[i].flag==False:
			points[i].flag=True
			cluster.pointids.append(i)
			ExpandCluster(points[i],cluster,points)
			
def Cluster(points,clusters):
	for pointid in points.keys():
		if points[pointid].flag==False:
			points[pointid].flag=True
			cluster = ClusterDef(len(clusters),[])
			cluster.pointids.append(pointid)
			ExpandCluster(points[pointid], cluster,points)
			clusters.append(cluster)	

def ComputeFmeasure(classes,clusters):
	#compute F-measure
	f=0
	n=0
	if n==0:
		return f
	for classname in classes.keys():
		group=classes[classname]
		m = 0
		for cluster in clusters:
			i = len(set(group) & set(cluster.pointids))
			v = float(2*i)/float((len(group) + len(cluster.pointids)))
			if m < v:
				m=v		
		n = n + len(group)
		f = f +	(len(group)*m)	
	return float(f)/float(n) 

def LoadClasses(allseqrecords,classificationfilename,pos):
	classificationfile= open(classificationfilename)
#	records= open(classificationfilename,errors='ignore')
	classification={}
	classes={}
	seqrecords={}
	for line in classificationfile:
		if line.startswith("#"):
			 continue
		words=line.split("\t")
		seqid= words[0].rstrip().replace(">","")
		if seqid in allseqrecords.keys():
			classname=""
			if pos < len(words):
				classname=words[pos].rstrip()
			if classname=="" or classname=="unidentified":	
				continue
			classification[seqid]=classname
			seqrecords[seqid]=allseqrecords[seqid]
			if classname in classes.keys():
				classes[classname].append(seqid)
			else:
				classes.setdefault(classname,[seqid]) 
	classificationfile.close()
	return seqrecords,classes,classification

def MergeComplexes(complexes,classes):
	removednames=[]
	complexnames=[]
	for comp in complexes:
		for classname in comp:
			if not (classname in complexnames):
				complexnames.append(classname)
	namepoints={}
	for classname in complexnames:
		namepoint = PointDef(classname,False,[],None)
		namepoints.setdefault(classname,namepoint)	
	#Load neighbor for points
	for comp in complexes:
		for classname in comp:
			for classname2 in comp:
				if not classname2 in namepoints[classname].neighbors:
					namepoints[classname].neighbors.append(classname2)			
	nameclusters=[]	
	Cluster(namepoints,nameclusters)	
	for namecluster in nameclusters:
		maxseqnumber=0
		ref=""
		for complexname in namecluster.pointids:
			seqnumber=len(classes[complexname])
			if seqnumber>=maxseqnumber:
				maxseqnumber=seqnumber
				ref=complexname
		for complexname in namecluster.pointids:
			if complexname!=ref:
				removednames.append(complexname)
	return removednames

def SaveClusters(clusters,seqrecords,classes,classification,output,outputfastafilename):
	outputfile=open(output,"w")
	outputfastafile=open(outputfastafilename,"w")
	outputfile.write("ClusterID\tSequenceID\tClassification\tPrediction\n")
	complexes=[]
	numberlist=[]
	for cluster in clusters:
		speciescomplex=[]
		seqnumbers=[]
		for seqid in cluster.pointids:
			seqrecord = seqrecords[seqid]
			classname=classification[seqid]
			if classname in speciescomplex:
				i=speciescomplex.index(classname)
				seqnumbers[i]=seqnumbers[i] +1 
			else:
				speciescomplex.append(classname)
				seqnumbers.append(1)
		complexes.append(speciescomplex)
		numberlist.append(seqnumbers)			
		maxindex=seqnumbers.index(max(seqnumbers))
		classname=speciescomplex[maxindex]
		for id in cluster.pointids:
			seqrecord = seqrecords[id]
			outputfile.write(str(cluster.id) + "\t" + seqrecord.description + "\t" + classification[id] + "\t" + classname + "\n")		
	outputfile.close()
	#remove complexes
	complexnames=MergeComplexes(complexes,classes)
	for seqid in seqrecords.keys():
		seqrecord=seqrecords[seqid]
		if not classification[seqid] in complexnames:
			outputfastafile.write(">"+seqrecord.description+"\n")
			outputfastafile.write(str(seqrecord.seq)+"\n")
	outputfastafile.close()	

def GetPosition(classificationfilename,rank):
	pos=-1
	classificationfile=open(classificationfilename)
	header=classificationfile.readline()
	header=header.rstrip()
	classificationfile.close()
	texts=header.split("\t")
	isError=False
	if rank in texts:
		pos=texts.index(rank)
	else:
		print("The rank " + rank + " is not given in the classification." )
		isError=True
	return pos,isError
	
if __name__ == "__main__":
	outputfastafilename=GetWorkingBase(fastafilename) + ".diff.fasta"	
	outputname=GetWorkingBase(fastafilename) + ".similar"
	allseqrecords=SeqIO.to_dict(SeqIO.parse(fastafilename, "fasta"))
	sys.setrecursionlimit(len(allseqrecords)*2)
	classificationpos,isError=GetPosition(classificationfilename,rank)
	if isError==True:
		sys.exit()
	seqrecords,classes,classification=LoadClasses(allseqrecords,classificationfilename,classificationpos)
	if len(seqrecords)==0:
		print("No classification names are available for the sequences at the position " + str(classificationpos) + str("."))
		os.sys.exit()
	#load similarity matrix
	if simfilename=="" or simfilename==None:
		simfilename=GetWorkingBase(fastafilename) + ".sim"
	simmatrix={}
	if os.path.exists(simfilename):
		print("Loading similarity matrix " + simfilename)
		simmatrix=LoadSim(simfilename)
	else:	
		print("Computing similarity matrix...")
		simmatrix=ComputeSim(fastafilename,seqrecords,mincoverage)
		print("Save similarity matrix " + simfilename)
		if len(seqrecords.keys())!=len(allseqrecords.keys()):
			simfilename=GetWorkingBase(simfilename) + "." + str(classificationpos) + ".sim"
		SaveSim(simmatrix,simfilename)	
	
	#load neighbors
	neighborlist = LoadNeighbors(seqrecords,simmatrix,threshold)
	print("Threshold\tFmeasure")
	points=LoadPoints(neighborlist,seqrecords)
	#cluster the sequences
	clusters=[]
	Cluster(points,clusters)
	fmeasure=ComputeFmeasure(classes,clusters)
	print(str(threshold) + "\t" + str(fmeasure))
	output=GetWorkingBase(fastafilename) + ".classified"
	SaveClusters(clusters,seqrecords,classes,classification,outputname,outputfastafilename)
	print("The remained sequences are saved in file: " + outputfastafilename )
	print("The clusters are saved in file: " + outputname )





#!/usr/bin/env python
# FILE: addclassificationtosequenceheaders.py
# AUTHOR: Duong Vu
# CREATE DATE: 07 June 2019
import sys, argparse
if sys.version_info[0] >= 3:
	unicode = str
import numpy as np
import os
from Bio import SeqIO
import multiprocessing
nproc=multiprocessing.cpu_count()

parser=argparse.ArgumentParser(prog='addclassificationtosequenceheaders.py',  
							   usage="%(prog)s [options] -i fastafile -c classificationfilename -o output",
							   description='''Script that adds taxonomic classification to sequence headers. ''',
							   epilog="""Written by Duong Vu duong.t.vu@gmail.com""",
   )

parser.add_argument('-i','--input', required=True, help='the fasta file to be clustered.')
parser.add_argument('-o','--out', help='The fasta filename with sequence headers containing classification.') 
parser.add_argument('-c','--classification', required=True, help='the classification file in tab. format.')
parser.add_argument('-p','--classificationposition', default="", help='the classification positions for the prediction.')


args=parser.parse_args()
fastafilename= args.input
classificationfilename=args.classification
classificationpos=args.classificationposition
output=args.out

#fastafilename=sys.argv[1]
#classificationfilename=sys.argv[2]
#output=sys.argv[3]

def GetBase(filename):
	return filename[:-(len(filename)-filename.rindex("."))]

def LoadClassification(classificationfilename,classificationpos):
	classificationfile= open(classificationfilename, errors="ignore")
	poslist=[]
	if "," in classificationpos:
		texts=classificationpos.split(",")
		for t in texts:
			poslist.append(int(t))
	elif classificationpos!="":
		poslist.append(int(classificationpos))	
	seqids=[]
	classifications=[]
	numberoffeatures=0
	for line in classificationfile:
		elements=line.rstrip().split("\t")
		seqid = elements[0].replace(">","").rstrip()
		seqids.append(seqid)
		classification=(line[line.index("\t")+1:])
		texts=classification.split("\t")
		classification=""
		if len(poslist) >0:		
			for pos in poslist:
				pos=pos-1
				text=""
				if pos < len(texts):
					text=texts[pos]
				if text=="":
					text="unidentified"	
				text=text.replace(" ","_")	
				classification=classification + text + "|"
		else:
			for text in texts:
				text=text.rstrip()
				if text=="":
					text="unidentified"
				text=text.replace(" ","_")		
				classification=classification + text + "|"
		classification=classification[:-1] 		
		classifications.append(classification)
		if numberoffeatures==0:
			numberoffeatures=classification.count("|")
		else:
			numberoffeatures=min(numberoffeatures,classification.count("|"))
	classificationfile.close()
	return seqids,classifications,numberoffeatures

#####main###

#seqrecords=list(SeqIO.parse(fastafilename, "fasta"))
seqids,classifications,numberoffeatures=LoadClassification(classificationfilename,classificationpos)
outputfile=open(output,"w")
fastafile=open(fastafilename)

for line in fastafile:
	if line.startswith(">"):
		seqid=line.rstrip().replace(">","")
		if "|" in seqid:
			seqid=seqid[0:seqid.index("|")]
		classification=""
		if seqid in seqids:
			classification=unicode(classifications[seqids.index(seqid)])
		else:
			classification="|" * numberoffeatures
		header=">" + seqid + "|" + classification + "\n"
		outputfile.write(header)	
	else:		
		outputfile.write(line)
	
outputfile.close()		
fastafile.close()



#!/usr/bin/env python
# FILE: evaluate.py
# AUTHOR: Duong Vu
# CREATE DATE: 07 June 2021
#from sklearn.model_selection import StratifiedKFold
#from sklearn.preprocessing import StandardScaler
import sys
if sys.version_info[0] >= 3:
	unicode = str
import os, argparse
from sklearn.metrics import precision_recall_fscore_support
from sklearn.metrics import matthews_corrcoef
from sklearn.metrics import confusion_matrix
from sklearn.metrics import accuracy_score

parser=argparse.ArgumentParser(prog='evaluate.py',  
							   usage="%(prog)s [options] -i predictionfile -c queryclassification -r referenceclassification",
							   description='''Script that visualizes using Krona and computes metrices for the classification results.''',
							   epilog="""Written by Duong Vu duong.t.vu@gmail.com""",
   )

parser.add_argument('-i','--input', required=True, help='the assigment/classified file')
parser.add_argument('-o','--out', default="dnabarcoder", help='The output folder.')
parser.add_argument('-c','--queryclassification', required=True, help='the given classification file of the query if exists, in tab. format to compute classification metrices.')
parser.add_argument('-r','--refclassification', required=True, help='the given classification file of the query if exists, in tab. format to compute classification metrices.')
parser.add_argument('-seqidpos','--sequenceidposition', type=int,default=0, help='the position of sequence id in the classification file.')
parser.add_argument('-givenlabelpos','--givenlabelposition', type=int,default=-1, help='the position of given labels in the prediction file.')
parser.add_argument('-predpos','--predictionposition', type=int,default=-1, help='the position of predicted labels in the prediction file.')
parser.add_argument('-rankpos','--rankposition', type=int,default=-1, help='the position of ranks in the prediction file.')


args=parser.parse_args()
predictionfilename=args.input
queryclassificationfilename=args.queryclassification
refclassificationfilename=args.refclassification


outputpath=args.out

if not os.path.exists(outputpath):
	os.system("mkdir " + outputpath)

def GetBase(filename):
	return filename[:-(len(filename)-filename.rindex("."))]
	
def GetWorkingBase(filename):
	basename=os.path.basename(filename)
	if "." in basename:
		basename=basename[:-(len(basename)-basename.rindex("."))] 
	path=outputpath + "/" + filename
	return path

def GetTaxonomicClassification(level,header,texts):
	classification=""
	p_s=len(texts)
	p_g=len(texts)
	p_f=len(texts)
	p_o=len(texts)
	p_c=len(texts)
	p_p=len(texts)
	p_k=len(texts)
	i=0
	for text in header.split("\t"):
		text=text.rstrip()
		if text.lower()=="species":
			p_s=i
		elif text.lower()=="genus":
			p_g=i	
		elif text.lower()=="family":
			p_f=i	
		elif text.lower()=="order":
			p_o=i	
		elif text.lower()=="class":
			p_c=i	
		elif text.lower()=="phylum":
			p_p=i	
		elif text.lower()=="kingdom":
			p_k=i	
		i=i+1 
	species="unidentified"
	genus="unidentified"
	family="unidentified"
	order="unidentified" 
	bioclass="unidentified"
	phylum="unidentified"
	kingdom="unidentified"
	if p_s< len(texts):
		species=texts[p_s].rstrip()
	if p_g< len(texts):
		genus=texts[p_g].rstrip()
	if p_f< len(texts):
		family=texts[p_f].rstrip()
	if p_o< len(texts):
		order=texts[p_o].rstrip()
	if p_c< len(texts):
		bioclass=texts[p_c].rstrip()
	if p_p< len(texts):
		phylum=texts[p_p].rstrip()
	if p_k< len(texts):
		kingdom=texts[p_k].rstrip()
	taxonname=""
	rank=""
	if level <7 and kingdom!="unidentified":
		taxonname=kingdom
		rank="kingdom"
		classification="k__" + kingdom +";p__unidentified;c__unidentified;o__unidentified;f__unidentified;g__unidentified;s__unidentified"	
	if level <6 and phylum!="unidentified":
		taxonname=phylum
		rank="phylum"
		classification="k__" + kingdom +";p__"+phylum +";c__unidentified;o__unidentified;f__unidentified;g__unidentified;s__unidentified"
	if level <5 and bioclass!="unidentified":
		taxonname=bioclass
		rank="class"
		classification="k__" + kingdom +";p__"+phylum +";c__"+bioclass+";o__unidentified;f__unidentified;g__unidentified;s__unidentified"
	if level <4 and order!="unidentified":
		taxonname=order
		rank="order"
		classification="k__" + kingdom +";p__"+phylum +";c__"+bioclass +";o__"+ order + ";f__unidentified;g__unidentified;s__unidentified"
	if level <3 and family!="unidentified":
		taxonname=family
		rank="family"
		classification="k__" + kingdom +";p__"+phylum +";c__"+bioclass +";o__"+ order+";f__"+family +";g__unidentified;s__unidentified"
	if level <2 and genus!="unidentified":
		taxonname=genus
		rank="genus"
		classification="k__" + kingdom +";p__"+phylum +";c__"+bioclass +";o__"+ order+";f__"+family + ";g__"+ genus + "s__unidentified"
	if level <1 and species!="unidentified":
		taxonname=species
		rank="species"
		classification="k__" + kingdom +";p__"+phylum +";c__"+bioclass +";o__"+ order+";f__"+family + ";g__"+ genus+";s__"+species 	
	return classification,taxonname,rank

def LoadClassification(classificationfilename):
	classificationdict={}
	if not os.path.exists(classificationfilename):
		return classificationdict
	classificationfile=open(classificationfilename)
	header=next(classificationfile)
	i=0
	p_seqid=0
	for text in header.split("\t"):
		text=text.rstrip().replace(" ","")
		if ("seqid" in text) or ("sequenceid" in text):
			p_seqid=i
		i=i+1	
	for line in classificationfile:
		texts=line.split("\t")
		seqid = texts[p_seqid].replace(">","").rstrip()
		classificationdict.setdefault(seqid,"")
		classification,taxonname,rank=GetTaxonomicClassification(0,header,texts)
		classificationdict[seqid]=classification
	classificationfile.close()	
	return classificationdict

def LoadTaxa(classificationfilename):
	taxa=[]
	classificationfile=open(classificationfilename)
	for line in classificationfile:
		line=line.rstrip()
		texts=line.split("\t")
		for text in texts:
			if text!="":
				taxa.append(text)
	classificationfile.close()			
	return taxa

def GetLevel(rank):
	level=-1
	if rank=="species":	
		level=6
	elif rank=="genus":	
		level=5
	elif rank=="family":	
		level=4
	elif rank=="order":	
		level=3
	elif rank=="class":	
		level=2
	elif rank=="phylum":
		level=1
	elif rank=="kingdom":
		level=0	
	return level

def LoadPrediction(predictionfilename,queryclassificationdict,outputname,reftaxa):
	if outputname!="":
		outputfile=open(outputname,"w")
	given_labels=[]
	pred_labels=[]
	p_seqid=args.sequenceidposition
	p_rank=args.rankposition
	p_givenlabel=args.givenlabelposition
	p_predlabel=args.predictionposition
	predictionfile= open(predictionfilename, "r")
	header=next(predictionfile)
	if p_givenlabel==-1 or p_predlabel==-1 or p_rank==-1:
		texts=header.split("\t")
		i=0
		for text in texts:
			text=text.rstrip()
			if "givenlabel" in text.lower().replace(" ",""):
				p_givenlabel=i
			if "prediction" in text.lower().replace(" ",""):
				p_predlabel=i	
			if "rank" in text.lower().replace(" ",""):
				p_rank=i		
			i=i+1	
	if outputname!="":
		outputfile.write(header)
	for line in predictionfile:
		texts=line.split("\t")
		seqid=texts[p_seqid]
		classname=texts[p_predlabel]
		if classname=="":
			classname="unidentified"
		if classname=="unidentified":
			continue
		givenlabel=texts[p_givenlabel]
		rank=texts[p_rank]
		level=GetLevel(rank)
		if seqid in queryclassificationdict.keys():
			queryclassification=queryclassificationdict[seqid]
			if level >=0:
				queryclassification=queryclassification.replace("k__","").replace("p__","").replace("c__","").replace("o__","").replace("f__","").replace("g__","").replace("s__","")
				givenlabel=queryclassification.split(";")[level]
		if givenlabel!="" and givenlabel!="unidentified":
			givenlabel=givenlabel.replace("_"," ")
			classname=classname.replace("_"," ")
			if outputname!="":
				texts[p_predlabel]=classname
				texts[p_givenlabel]=givenlabel
				newline=""
				for text in texts:
					newline=newline + text + "\t"
				newline=newline[:-1]
				outputfile.write(newline)
			given_labels.append(givenlabel)
			pred_labels.append(classname)	
		classification=texts[3]
		classification=classification.replace("k__","").replace("p__","").replace("c__","").replace("o__","").replace("f__","").replace("g__","").replace("s__","")

	if outputname!="":
		outputfile.close()
	return given_labels,pred_labels

def CalculateMetrics(test_labels,pred_labels,labels): 
	if len(test_labels)==0:
		return 0,0,0,0,[],[],[],0,[]
	accuracy=accuracy_score(test_labels,pred_labels)
	precision,recall,fscore,support=precision_recall_fscore_support(test_labels,pred_labels,average='macro')
	precisionvector,recallvector,fscorevector,support=precision_recall_fscore_support(test_labels,pred_labels,labels=labels)
	mcc=matthews_corrcoef(test_labels,pred_labels)
	#cohenkappascore=cohen_kappa_score(test_labels,pred_labels,labels=labels)
	#print("Cohenkappascore "+ str(cohenkappascore))
	confusionmatrix=confusion_matrix(test_labels,pred_labels,labels=labels)
	return round(accuracy,4),round(precision,4),round(recall,4),round(fscore,4),precisionvector,recallvector,fscorevector,round(mcc,4),confusionmatrix
	
def CalculateClassificationMetrics(givenlabels,predlabels,reftaxa,reportname):
	tmp_givenlabels=[]
	tmp_predlabels=[]
	numberofspeciescomplexes=0
	i=0
	for label in givenlabels:
		genus=""
		predgenus=""
		predlabel=predlabels[i]
		if " " in label:
			genus=label.split(" ")[0]
			label=label.split(" ")[1]
		if " " in predlabels[i]:
			predgenus=predlabel.split(" ")[0]	
			predlabel=predlabel.split(" ")[1]
		if predgenus==genus and genus!="" and label!=predlabel:
			numberofspeciescomplexes=numberofspeciescomplexes+1
		tmp_givenlabels.append(label)		
		tmp_predlabels.append(predlabel)
		i=i+1		
	accuracy,precision,recall,fscore,precisionvector,recallvector,fscorevector,mcc,confusionmatrix=CalculateMetrics(tmp_givenlabels,tmp_predlabels,tmp_predlabels)
	filteredgivenlabels=[]
	filteredpredlabels=[]
	filterednumberofspeciescomplexes=0
	i=0
	for label in givenlabels:
		if label in reftaxa:
			genus=""
			predgenus=""
			predlabel=predlabels[i]
			if " " in label:
				genus=label.split(" ")[0]
				label=label.split(" ")[1]
			if " " in predlabels[i]:
				predgenus=predlabel.split(" ")[0]	
				predlabel=predlabel.split(" ")[1]
			if predgenus==genus and genus!="" and label!=predlabel:
				filterednumberofspeciescomplexes=filterednumberofspeciescomplexes+1	
			filteredgivenlabels.append(label)
			filteredpredlabels.append(predlabel)
		i=i+1			
	filteredaccuracy,filteredprecision,filteredrecall,filteredfscore,filteredprecisionvector,filteredrecallvector,filterdfscorevector,filteredmcc,filteredconfusionmatrix=CalculateMetrics(filteredgivenlabels,filteredpredlabels,filteredpredlabels)
	if numberofspeciescomplexes>0:
		print("Number of species complexes: " + str(numberofspeciescomplexes) + "(" +  str(round(numberofspeciescomplexes*100/len(given_labels),2)) + "%).")
	if filterednumberofspeciescomplexes>0:
		print("Number of species complexes that are present in the reference dataset: " + str(filterednumberofspeciescomplexes) + "(" +  str(round(filterednumberofspeciescomplexes*100/len(filteredgivenlabels),2)) + "%).")	
	print("Number of sequences with a given label: " + str(len(given_labels)) + "\t" + str(len(filteredgivenlabels)))
	print("The mcc of assigning the sequences with a given label: " + str(mcc) + "\t" + str(filteredmcc))
	print("The accuracy of assigning the sequence with a given label: " + str(accuracy) + "\t" + str(filteredaccuracy))
	print("The precision of assigning the sequence with a given label: " + str(precision) + "\t" + str(filteredprecision))
	print("The fscore of assigning the sequence with a given label: " + str(fscore) + "\t" + str(filteredfscore))
		
	report=open(reportname,"w")
	report.write("Number of sequences with a given label\tMcc\tAccuracy\tPrecision\tFscore\tNumber of sequences with a given label present in the references\tMcc\tAccuracy\tPrecision\tFscore\n")
	report.write(str(len(givenlabels))+ "\t" + str(mcc) + "\t" + str(accuracy) + "\t" + str(precision) + "\t" + str(fscore) + "\t")
	report.write(str(len(filteredgivenlabels))+ "\t" + str(filteredmcc) + "\t" + str(filteredaccuracy) + "\t" + str(filteredprecision) + "\t" + str(filteredfscore) + "\n")
	
	report.write("Taxonname\tAccuracy\tPrecision\tFscore\n")
	i=0
	existinglabels=[]
	for label in pred_labels:
		if label in existinglabels:
			i=i+1
			continue
		existinglabels.append(label)
		report.write(label + "\t" + str(recallvector[i]) + "\t" + str(precisionvector[i]) + "\t" + str(fscorevector[i]) + "\n")
		i=i+1
	report.close()
	print("Accuracy, precision, and fscore of the taxa are given in file " + reportname +  ".")

if __name__ == "__main__":
	queryclassificationdict=LoadClassification(queryclassificationfilename)
	outputname=GetBase(predictionfilename) + ".labeled"
	reftaxa=LoadTaxa(refclassificationfilename)	
	given_labels,pred_labels= LoadPrediction(predictionfilename,queryclassificationdict,outputname,reftaxa)
	if outputname!="":
		print("The assigned sequences with given labels are saved in file " + outputname + ".") 
	if len(given_labels) >0:
		reportname=GetBase(outputname) + ".report"
		CalculateClassificationMetrics(given_labels,pred_labels,reftaxa,reportname)
	


#!/usr/bin/env python
# FILE: cleansequenceheaders.py, requiring Python 2.7 if classification file contains non-ascii characters
# AUTHOR: Duong Vu
# CREATE DATE: 07 June 2021
import sys, argparse
import numpy as np
import os
from Bio import SeqIO
import json
import multiprocessing
nproc=multiprocessing.cpu_count()
#from keras.utils import np_utils

parser=argparse.ArgumentParser(prog='cleansequenceheaders.py',  
							   usage="%(prog)s [options] -i fastafile -o output",
							   description='''Script that keeps unique sequence id from sequence headers. ''',
							   epilog="""Written by Duong Vu duong.t.vu@gmail.com""",
   )

parser.add_argument('-i','--input', required=True, help='the fasta file.')
parser.add_argument('-o','--out', help='The output fasta file with sequence headers being cleaned. Only the first part of the sequence headers separated by "|" will be kept.')

args=parser.parse_args()
fastafilename= args.input
output=args.out

#fastafilename=sys.argv[1]
#text=sys.argv[2]
#output=sys.argv[3]

outputfile=open(output,"w")
fastafile=open(fastafilename)

seqids=[]
counts=[]
for line in fastafile:
	if line.startswith(">"):
		header=line.rstrip()
		seqid=header
		if "|" in header:
			seqid=header.split("|")[0]	
		if seqid in seqids:
			counts[seqids.index(seqid)]=counts[seqids.index(seqid)]+1
			seqid=seqid + "_" + str(counts[seqids.index(seqid)])
		else:	
			seqids.append(seqid)
			counts.append(1)
		outputfile.write(seqid + "\n")	
	else:		
		outputfile.write(line)
fastafile.close()
outputfile.close()		
i=0
for seqid in seqids:
	if counts[i] >1:
		print("Sequence " + seqid.replace(">","") + " has " + str(counts[i]) + " sequences.")
	i=i+1




#!/usr/bin/env python
# FILE: addclassificationtosequenceheaders.py
# AUTHOR: Duong Vu
# CREATE DATE: 07 June 2019
import sys, argparse
if sys.version_info[0] >= 3:
	unicode = str
import numpy as np
import os
from Bio import SeqIO
import multiprocessing
nproc=multiprocessing.cpu_count()

parser=argparse.ArgumentParser(prog='tab2fasta.py',  
							   usage="%(prog)s [options] -i tabfile -o output",
							   description='''Script that creates a fasta file from a tab. file. ''',
							   epilog="""Written by Duong Vu duong.t.vu@gmail.com""",
   )

parser.add_argument('-i','--input', required=True, help='the tab. file.')
parser.add_argument('-o','--out', help='The fasta filename.') 
parser.add_argument('-seqidpos','--seqidpos', required=True, help='The position of seqid in the tab. file.') 
parser.add_argument('-seqpos','--seqpos', type=int, required=True, help='The position of sequences in the tab. file.') 


args=parser.parse_args()
tabfilename= args.input
seqidpos=args.seqidpos
seqpos=args.seqpos
output=args.out

#fastafilename=sys.argv[1]
#classificationfilename=sys.argv[2]
#output=sys.argv[3]

def GetBase(filename):
	return filename[:-(len(filename)-filename.rindex("."))]

seqidposlist=[]
if "," in seqidpos:
	texts=seqidpos.split(",")
	for text in texts:
		seqidposlist.append(int(text))
else:
	seqidposlist.append(int(seqidpos))		

outputfile=open(output,"w")
tabfile=open(tabfilename, errors='ignore')
next(tabfile)
seqids=[]
for line in tabfile:
	#line=unicode(line,'latin1')
	texts=line.split("\t")
	seqid=""
	for pos in seqidposlist:
		seqid=texts[pos]
		if seqid!="":
			break
	if seqid =="":
		print(line)
		continue
	seq=texts[seqpos]
	if seq=="" or "fail" in seq:
		print(line)
		continue 
	if seqid in seqids:
		print(line)
		continue
	seqids.append(seqid)
	outputfile.write(">" + seqid + "\n")
	outputfile.write(seq.replace("\n","") + "\n")
outputfile.close()		
tabfile.close()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# FILE: predictlocally.py
# AUTHOR: Duong Vu
# CREATE DATE: 08 oct 2020

import os, argparse
import sys
from Bio import SeqIO
import matplotlib.pyplot as plt
plt.rc('font',size=6)
#from matplotlib.patches import Polygon
import numpy as np
import multiprocessing
import json

parser=argparse.ArgumentParser(prog='predict.py', 
							   usage="%(prog)s [options] -i fastafile -c classificationfile -p classificationposition -st startingthreshold -et endthreshold -s step -ml minalignmentlength",
							   description='''Script that predicts an optimal threshold to separate the sequences based on the given classification''',
							   epilog="""Written by Duong Vu duong.t.vu@gmail.com""",
   )

parser.add_argument('-i','--input', required=True, help='the fasta file')
parser.add_argument('-o','--out', default="dnabarcoder", help='The output folder.')
parser.add_argument('-prefix','--prefix', default="", help='the prefix of output filenames.')
parser.add_argument('-label','--label',default="", help='The label to display in the figure.')
parser.add_argument('-labelstyle','--labelstyle', default='normal', help='The label style to be displayed: normal, italic, or bold.')
parser.add_argument('-c','--classification', default="", required=True, help='the classification file in tab. format.')
#parser.add_argument('-p','--classificationpositions', default="", help='the classification positions for the prediction, separated by ",".')
parser.add_argument('-ranks','--classificationranks', default="", help='the classification ranks for the prediction, separated by ",".')
parser.add_argument('-st','--startingthreshold', type=float, default=0, help='starting threshold')
parser.add_argument('-et','--endthreshold', type=float, default=0, help='ending threshold')
parser.add_argument('-s','--step', type=float, default=0.001, help='the step to be increased for the threshold after each step of the prediction.')
parser.add_argument('-ml','--minalignmentlength', type=int,default=400, help='Minimum sequence alignment length required for BLAST. For short barcode sequences like ITS2 (ITS1) sequences, minalignmentlength should probably be set to smaller, 50 for instance.')
parser.add_argument('-sim','--simfilename', help='The similarity matrix of the sequences if exists.')
#parser.add_argument('-hp','--higherclassificationpositions', default="", help='The prediction is based on the whole dataset if hp="". Otherwise it will be predicted based on different datasets obtained at the higher classifications, separated by ",".')
parser.add_argument('-higherranks','--higherclassificationranks', default="", help='The prediction is done on the whole dataset if higherranks="". Otherwise it will be predicted for different datasets obtained at the higher classifications, separated by ",".')
parser.add_argument('-mingroupno','--mingroupno', type=int, default=10, help='The minimum number of groups needed for prediction.')
parser.add_argument('-minseqno','--minseqno', type=int, default=30, help='The minimum number of sequences needed for prediction.')
parser.add_argument('-maxsimmatrixsize','--maxSimMatrixSize', type=int, default=20000, help='The maximum number of sequences to load or compute a full similarity matrix. In case the number of sequences is greater than this number, only similarity values greater than 0 will be loaded to avoid memory problems.')
parser.add_argument('-taxa','--taxa', default="", help='The selected taxa separated by commas for local prediction. If taxa=="", all the clades at the given higher positions are selected for prediction.')
parser.add_argument('-redo','--redo', default="", help='Recompute F-measure for the current parameters.')


args=parser.parse_args()
fastafilename= args.input
classificationfilename=args.classification
#classificationpos=args.classificationpositions
#higherclassificationpos=args.higherclassificationpositions
classificationranks=args.classificationranks
higherclassificationranks=args.higherclassificationranks
threshold = args.startingthreshold
endthreshold=args.endthreshold
step=args.step
mincoverage=args.minalignmentlength
outputfolder=args.out
simfilename=args.simfilename
minGroupNo=args.mingroupno
minSeqNo=args.minseqno
#taxa=args.taxa
prefix=args.prefix
label=args.label
outputpath=args.out
redo=args.redo

if not os.path.exists(outputpath):
	os.system("mkdir " + outputpath)


nproc=multiprocessing.cpu_count()

def GetBase(filename):
	return filename[:-(len(filename)-filename.rindex("."))]  
	
#def GetWorkingBase(filename):
#	basename=os.path.basename(filename)
#	if "." in basename:
#		basename=basename[:-(len(basename)-basename.rindex("."))] 
#	path=outputpath + "/" + basename
#	return path
	
def GetWorkingBase(filename): 
	path=outputpath + "/" + filename
	return path

class PointDef:
	def __init__(self, id,flag,neighbors,seqrecord):
		self.id=id
		self.flag=flag
		self.neighbors = neighbors
		self.seqrecord=seqrecord

class ClusterDef:
	def __init__(self, id, pointids):
		self.id=id
		self.pointids=pointids

def GetSeqIndex(seqname,seqlist):
	i=0
	for seq in seqlist:
		if (seqname == seq.id):
			return i
		i = i + 1
	return -1
	
def LoadSim(simfilename):
	simmatrix = {} #we use dictionary to reduce the memory constraints 
	simfile = open(simfilename)
	seqids=[]
	for line in simfile:
		numbers=line.rstrip().split(" ")
		i=numbers[0]
		j=numbers[1]
		seqids.append(i)
		seqids.append(j)
		if i not in simmatrix.keys():
			simmatrix.setdefault(i, {})
		simmatrix[i][j]=float(numbers[2])
	seqids=list(set(seqids))	
	for seqid1 in seqids:
		if not (seqid1 in simmatrix.keys()):
			simmatrix.setdefault(seqid1, {})
			simmatrix[seqid1][seqid1]=1	
		if len(seqids) < args.maxSimMatrixSize: #load full matrix	
			for seqid2 in seqids:
				if not seqid2 in simmatrix[seqid1].keys():
					#simmatrix[seqid1].setdefault(seqid2,0)
					simmatrix[seqid1][seqid2]=0
	simfile.close()		
	return simmatrix

def SaveSim(simmatrix,simfilename):
	simfile=open(simfilename,"w")
	for i in simmatrix.keys():
		for j in simmatrix[i].keys():
			simfile.write(str(i) + " " + str(j) + " " + str(simmatrix[i][j]) + "\n")
	simfile.close()
	
def ComputeSim(fastafilename,seqrecords,mincoverage):
	blastoutput = fastafilename + ".blast.out"		
	blastdb=fastafilename + ".db"		
	#blast
	print("Comparing the sequences of " + fastafilename + " using Blast...")
	makedbcommand = "makeblastdb -in " + fastafilename + " -dbtype \'nucl\' " +  " -out " + blastdb
	print(makedbcommand)
	os.system(makedbcommand)
	blastcommand = "blastn -query " + fastafilename + " -db  " + blastdb + " -task blastn-short -outfmt 6 -out " + blastoutput + " -num_threads " + str(nproc)
	if mincoverage >=400:
		blastcommand = "blastn -query " + fastafilename + " -db " + blastdb + " -outfmt 6 -out " + blastoutput + " -num_threads " + str(nproc)
	print(blastcommand)
	os.system(blastcommand)
	if not os.path.exists(blastoutput):
		print("Cannot compare the sequences of " + fastafilename + " using Blast...")
		logfile=open(GetWorkingBase((os.path.basename(args.input))) + ".predict.log","w")
		logfile.write("Cannot compare the sequences of " + fastafilename + " using Blast...")
		logfile.write("Make BLAST database command: " + makedbcommand + "\n")
		logfile.write("No output for the BLAST command: " + blastcommand + "\n")
		logfile.write("Please rerun prediction for " + os.path.basename(fastafilename) + ".")
		logfile.close()
		return {}
	print("Reading Blast results of " + fastafilename + "...")
	simmatrix={}
	for seqid1 in seqrecords.keys():
		simmatrix.setdefault(seqid1,{})
		for seqid2 in seqrecords.keys():
			if seqid1==seqid2:
				simmatrix[seqid1][seqid2]=1
			elif len(seqrecords.keys()) < args.maxSimMatrixSize: #load full matrix	
				simmatrix[seqid1][seqid2]=0
	#read blast output
	blastoutputfile = open(blastoutput)
	score=0
	for line in blastoutputfile:
		if line.rstrip()=="":
			continue
		words = line.split("\t")
		i = words[0].rstrip()
		j = words[1].rstrip()
		pos1 = int(words[6])
		pos2 = int(words[7])
		iden = float(words[2]) 
		sim=float(iden)/100
		coverage=abs(pos2-pos1)
		score=sim
		if coverage < mincoverage:
			score=float(score * coverage)/mincoverage
		if len(seqrecords.keys()) < args.maxSimMatrixSize: #the full sim matrix has been loaded 
			if simmatrix[i][j] < score:
				simmatrix[i][j]=round(score,4)
				simmatrix[j][i]=round(score,4)
		else:		
			if j in simmatrix[i].keys():
				if simmatrix[i][j] < score:
					simmatrix[i][j]=round(score,4)
					simmatrix[j][i]=round(score,4)
			else:
				simmatrix[i][j]=round(score,4)
				simmatrix[j][i]=round(score,4)	
		#simmatrix[j][i]=score
	os.system("rm " + blastoutput)
	os.system("rm " + blastdb + "*")
	#os.system("rm " + blastdb + ".*")
	return simmatrix

def LoadNeighbors(seqids,subsimmatrix,threshold):
	neighbordict={}
	for seqid in seqids:
		neighbordict.setdefault(seqid, [])
	if len(subsimmatrix.keys()) < args.maxSimMatrixSize:	 #the full matrix has been loaded
		for i in seqids:
			for j in seqids:
				if subsimmatrix[i][j] >= threshold:
					neighbordict[i].append(j)
					neighbordict[j].append(i)
	else:
		for i in seqids:
			for j in seqids:
				if j in subsimmatrix[i].keys():
					if subsimmatrix[i][j] >= threshold:
						neighbordict[i].append(j)
						neighbordict[j].append(i)					
	#os.system("rm out.txt")
	return neighbordict

def LoadPoints(neigbordict,seqrecords):
	points={}
	for seqid in  seqrecords.keys():
		point = PointDef(seqid,False,neigbordict[seqid],seqrecords[seqid])
		points[seqid]=point
	return points

def ExpandCluster(root,cluster,points):
	for i in root.neighbors:
		if points[i].flag==False:
			points[i].flag=True
			cluster.pointids.append(i)
			ExpandCluster(points[i],cluster,points)		

def Cluster(points,clusters):
	for pointid in points.keys():
		if points[pointid].flag==False:
			points[pointid].flag=True
			cluster = ClusterDef(len(clusters),[])
			cluster.pointids.append(pointid)
			ExpandCluster(points[pointid], cluster,points)
			clusters.append(cluster)
			
def ComputeFmeasure(classes,clusters):
	#compute F-measure
	f=0
	n=0
	for classname in classes.keys():
		group=classes[classname]
		m = 0
		for cluster in clusters:
			i = len(set(group) & set(cluster.pointids))
			v = float(2*i)/float((len(group) + len(cluster.pointids)))
			if m < v:
				m=v		
		n = n + len(group)
		f = f +	(len(group)*m)	
	return float(f)/float(n) 

def LoadClasses(seqids,classificationfilename,pos):
	classificationfile= open(classificationfilename)
#	records= open(classificationfilename,errors='ignore')
	classification={}
	classes={}
	for line in classificationfile:
		if line.startswith("#"):
			 continue
		words=line.split("\t")
		seqid= words[0].rstrip().replace(">","")
		if seqid in seqids:
			classname=""
			if pos < len(words):
				classname=words[pos].rstrip()
			classification[seqid]=classname
			if classname in classes.keys():
				classes[classname].append(seqid)
			else:
				classes.setdefault(classname,[seqid]) 
	classificationfile.close()
	return classes,classification

def isfloat(value):
  try:
    float(value)
    return True
  except ValueError:
    return False

def ComputeSubSim(datasetname,records,simmatrix):
#	if datasetname=="All":
#		return simmatrix
#	subsimmatrix={}
	if simmatrix!={}:
#		for seqid1 in records.keys():
#			subsimmatrix.setdefault(seqid1,{})
#			for seqid2 in records.keys():
#				subsimmatrix[seqid1][seqid2]=simmatrix[seqid1][seqid2]	
		return simmatrix
	else:
		#save sequence records to a fasta file
		subfastafilename=GetWorkingBase(datasetname) + ".fasta"
		SeqIO.write(records.values(), subfastafilename, "fasta")			
		subsimmatrix=ComputeSim(subfastafilename,records,mincoverage)
		if subsimmatrix!={}:
			os.system("rm " + subfastafilename)
	return subsimmatrix
	
def Predict(datasetname,prediction_datasetname,records,classes,simmatrix):
	thresholds=[]
	fmeasures=[]	
	t=round(threshold,4)
	optthreshold=0
	bestFmeasure=0
	fmeasuredict={}
	if 'cut-off' in prediction_datasetname.keys():	
		optthreshold=prediction_datasetname['cut-off']
	if 'confidence' in prediction_datasetname.keys():		
		bestFmeasure=prediction_datasetname['confidence']
	if 'fmeasures' in prediction_datasetname.keys():			
		fmeasuredict=prediction_datasetname['fmeasures']
	isError=False		
	subsimmatrix={}	
	#compute optimal threshold
	while t <= endthreshold:
		print("Computing F-measure for threshold " + str(t))
		fmeasure=0
		#if str(t) in fmeasuredict.keys() and datasetdict['fasta filename']==fastafilename and datasetdict['classification filename']==classificationfilename:
		if str(t) in fmeasuredict.keys() and args.redo=="":
			fmeasure=fmeasuredict[str(t)]
		else:	
			if subsimmatrix=={}:
				#compute sub simmatrix
				subsimmatrix=ComputeSubSim(datasetname,records,simmatrix)
				if subsimmatrix=={}:
					isError=True	
					break
			#compute fmeasure
			neighbordict = LoadNeighbors(records.keys(),subsimmatrix,t)	
			points=LoadPoints(neighbordict,records)
			clusters=[]
			Cluster(points,clusters)	
			fmeasure=round(ComputeFmeasure(classes,clusters),4)
			fmeasuredict[str(t)]=fmeasure
		if fmeasure > bestFmeasure or (fmeasure==bestFmeasure and optthreshold >t) :
			bestFmeasure=fmeasure
			optthreshold=t
		thresholds.append(t)
		fmeasures.append(fmeasure)
		t=round(t+step,4)
		print("F-measure: " + str(fmeasure))
	if isError==False:	
		prediction_datasetname['cut-off']=optthreshold
		prediction_datasetname['confidence']=bestFmeasure
		#prediction_datasetname['min alignment length']=mincoverage
		prediction_datasetname['sequence number']=len(records)
		prediction_datasetname['group number']=len(classes)
		prediction_datasetname['fmeasures']=fmeasuredict
	return thresholds,fmeasures,optthreshold,bestFmeasure,isError

def GetPositionList(classificationfilename,ranklist,higherranklist):
	positionlist=[]
	higherpositionlist=[]
	classificationfile=open(classificationfilename)
	header=classificationfile.readline()
	header=header.rstrip()
	classificationfile.close()
	texts=header.split("\t")
	isError=False
	for rank in ranklist:
		if rank in texts:
			pos=texts.index(rank)
			positionlist.append(pos)
		else:
			print("The rank " + rank + " is not given in the classification." )
			isError=True
	if isError==False:
		for rank in higherranklist:
			if rank in texts:
				pos=texts.index(rank)
				higherpositionlist.append(pos)
			else:
				print("The higher rank " + rank + " is not given in the classification." )
				isError=True
	return positionlist,higherpositionlist,isError
	
def GenerateDatasetsForPrediction(seqrecords,classificationfilename,pos,higherpos,taxa):
	taxalist=[]
	if "," in taxa:
		taxalist=taxa.split(",")
	elif taxa!="":
		taxalist.append(taxa)
	datasets={}
	classificationfile= open(classificationfilename)
	for line in classificationfile:
		if line.startswith("#"):
			continue 		
		texts=line.rstrip().split("\t")
		seqid=texts[0].replace(">","")
		classname=""
		higherclassname=""
		if pos < len(texts):
			classname=texts[pos].rstrip()
		if higherpos > 0:
			if higherpos < len(texts):
				higherclassname=texts[higherpos].rstrip()
		if (len(taxalist) >0) and higherclassname!="" and not (higherclassname in taxalist) :
			continue
		if classname != "" and higherclassname !="" and classname != "unidentified" and higherclassname !="unidentified" and seqid in seqrecords.keys():
			if not higherclassname in datasets.keys():
					datasets.setdefault(higherclassname,{})
			datasets[higherclassname][seqid]=seqrecords[seqid]	
	classificationfile.close()
	return datasets

def GenerateDatasets(seqrecords,classificationfilename,pos,higherpositionlist,taxa):
	alldatasets={}
	if len(higherpositionlist) !=0:		
		for higherpos in higherpositionlist:
			datasets=GenerateDatasetsForPrediction(seqrecords,classificationfilename,pos,higherpos,taxa)
			alldatasets.update(datasets)
		return alldatasets	
	else:
		alldatasets.setdefault("All",{})
		classificationfile= open(classificationfilename)
		for line in classificationfile:
			texts=line.rstrip().split("\t")
			seqid=texts[0].replace(">","")
			classname=""
			if pos < len(texts):
				classname=texts[pos].rstrip()	
			if classname != "" and classname != "unidentified" and seqid in seqrecords.keys():
				alldatasets["All"][seqid]=seqrecords[seqid]
		classificationfile.close()
	return alldatasets

def LoadPrediction(predictionfilename):
	existingprediction={}
	#load classes
	with open(predictionfilename,encoding='latin1') as json_file:
		existingprediction = json.load(json_file)
	return existingprediction
			
def LoadPredictionAtPos(prediction_datasetname):
	thresholds=[]
	fmeasures=[]	
	optthreshold=0
	if 'cut-off' in prediction_datasetname.keys():
		optthreshold=prediction_datasetname['cut-off']
	bestFmeasure=0
	if 'confidence' in prediction_datasetname.keys():
		bestFmeasure=prediction_datasetname['confidence']
	seqno=0
	if 'sequence number' in prediction_datasetname.keys():
		seqno=prediction_datasetname['sequence number']
	groupno=0	
	if 'group number' in prediction_datasetname.keys():
		groupno=prediction_datasetname['group number']
	minalignmentlength=0	
	if 'min alignment length' in prediction_datasetname.keys():	
		minalignmentlength=prediction_datasetname['min alignment length']	
	fmeasuredict={}
	if 'fmeasures' in prediction_datasetname.keys():
		fmeasuredict=prediction_datasetname['fmeasures']
	for t in fmeasuredict.keys():
		if float(t)>=args.startingthreshold and ((float(t)<=args.endthreshold and args.endthreshold>0) or args.endthreshold==0):
			thresholds.append(float(t))
			fmeasures.append(fmeasuredict[t])
	#sorting
	keydict = dict(zip(fmeasures,thresholds))
	fmeasures.sort(key=keydict.get)
	thresholds.sort()
	return thresholds,fmeasures,optthreshold,bestFmeasure,seqno,groupno,minalignmentlength

def SavePrediction(predictiondict,outputnamewithfmeasures,outputnamewithoutfmeasures):
	#save the whole prediction file
	with open(outputnamewithfmeasures,"w") as json_file:
		if sys.version_info[0] >= 3:
			json.dump(predictiondict,json_file,indent=2)
		else:
			json.dump(predictiondict,json_file,encoding='latin1',indent=2)
	#save only the final results (without fmeasures)
	finalresults=predictiondict.copy()
	for rank in finalresults.keys():
		datasets=finalresults[rank]
		datasetnames=list(datasets.keys()).copy()
		for datasetname in datasetnames:
			dataset=datasets[datasetname]
			if "fmeasures" in dataset.keys():
				del dataset["fmeasures"]
			seqno=0 
			if "sequence number" in dataset.keys():
				seqno=dataset["sequence number"]
			groupno=0
			if "group number" in dataset.keys():
				groupno=dataset["group number"]
			minalignmentlength=0
			if "min alignment length" in dataset.keys():
				minalignmentlength=dataset["min alignment length"]
			if groupno < minGroupNo or seqno < minSeqNo or (minalignmentlength>0 and minalignmentlength <mincoverage):	#delete the cutoffs that dont have enough sequences and groups for prediction
				del datasets[datasetname]			
	with open(outputnamewithoutfmeasures,"w") as json_file:
		if sys.version_info[0] >= 3:
			json.dump(finalresults,json_file,indent=2)
		else:
			json.dump(finalresults,json_file,encoding='latin1',indent=2)
	#save as tab. format
	textoutput=outputnamewithoutfmeasures+".txt"
	textfile=open(textoutput,"w")
	textfile.write("Rank\tDataset\tcut-off\tconfidence\tsequence number\tgroup number\tmin alignment length\tfasta filename\tclassification filename\n")
	for rank in finalresults.keys():
		datasets=finalresults[rank]
		for datasetname in datasets.keys():
			dataset=datasets[datasetname]
			cutoff=0 
			if "cut-off" in dataset.keys():
				cutoff=dataset["cut-off"]
			confidence=0 
			if "confidence" in dataset.keys():
				confidence=dataset["confidence"]	
			seqno=0 
			if "sequence number" in dataset.keys():
				seqno=dataset["sequence number"]
			groupno=0
			if "group number" in dataset.keys():
				groupno=dataset["group number"]
			minalignmentlength=0
			if "min alignment length" in dataset.keys():
				minalignmentlength=dataset["min alignment length"]
			fastafilename=""
			if "fasta filename" in dataset.keys():
				fastafilename=dataset["fasta filename"]
			classificationfilename=""
			if "classification filename" in dataset.keys():	
				classificationfilename=dataset["classification filename"]
			#classificationposition=dataset["classification position"]
			textfile.write(rank+"\t" + datasetname + "\t"+str(cutoff)+"\t"+str(confidence)+"\t"+str(seqno)+"\t"+str(groupno)+"\t"+ str(minalignmentlength) + "\t" + fastafilename+"\t"+classificationfilename+"\n")
	textfile.close()
			

def GetTaxa(classificationfilename,pos,highertaxon):
	taxa=[]
	classificationfile=open(classificationfilename)
	for line in classificationfile:
		if line.startswith("#"):
			continue		
		texts=line.rstrip().split("\t")
		if not highertaxon in texts:
			continue
		if pos<len(texts):
			taxon=texts[pos]
			if not taxon in taxa:
				taxa.append(taxon)
	classificationfile.close()
	return taxa
	
def PlotPrediction(datasetname,thresholdlist,fmeasurelist,optthresholds,bestFmeasures,features,datasetnames,figoutput):
	if len(thresholdlist) >5:
		fig, ax = plt.subplots(figsize=(6,3))
	else:	
		fig, ax = plt.subplots(figsize=(4,3)) 
	ax.set_xlabel("Cut-off")
	ax.set_ylabel('F-measure')
	colors = plt.cm.rainbow(np.linspace(0, 1, len(thresholdlist)))
	labels=[]
	i=0
	for thresholds in thresholdlist:
		fmeasures=fmeasurelist[i]
		optthreshold=optthresholds[i]
		bestFmeasure=bestFmeasures[i]
		ax.plot(np.array(thresholds), np.array(fmeasures),color=colors[i])
		labels.append(features[i] + " cut-off for " + datasetnames[i] + ": "  + str(round(optthresholds[i],4)))
		ax.text(round(optthreshold,4), 0.97, round(bestFmeasure,4), transform=ax.get_xaxis_transform(), horizontalalignment='center', size='x-small', color=colors[i])
		i=i+1
	ax.set_title(datasetname + ": predicting similarity cut-offs for sequence identification")	
	plt.legend(labels, loc="lower left", bbox_transform=plt.gcf().transFigure,prop={'style': args.labelstyle})
	plt.tight_layout()
	plt.savefig(figoutput, dpi = 500)
	plt.show()	
	
def PlotResults(prefix,optthresholds,bestFmeasures,features,datasetnames,localfigoutput):
	#sort all according to increasing order of optthresholds
	if len(optthresholds)==0:
		return
	optthresholds,bestFmeasures,datasetnames,features = zip(*sorted(zip(optthresholds,bestFmeasures,datasetnames,features)))
	labels=[]
	i=0
	for datasetname in datasetnames:
		if len(set(features))==1:
			labels.append(datasetnames[i] )
		else:
			labels.append(features[i] + " cut-off for " + datasetnames[i] )
		i=i+1
	x = np.arange(len(labels))  # the label locations
	#width = 0.35  # the width of the bars
	fig, ax = plt.subplots(figsize=(6,3)) 
	ax2=ax.twinx()
	#ax.bar(x,optthresholds,color='b')
	ax.plot(np.array(labels), np.array(optthresholds), color='b')
	ax2.plot(np.array(labels), np.array(bestFmeasures), color='r')
	ax.set_ylabel('Cut-off',color='b')
	ax2.set_ylabel("F-measure",color="r")
	if len(set(features))==1:
		ax.set_title(prefix + ": " + features[0] + " cut-offs and F-measures predicted for different groups")
	else:
		ax.set_title(prefix + ": cut-offs and F-measures predicted for different groups")
	
	ax.set_xticks(x)
	ax.set_xticklabels(labels,rotation=90, style=args.labelstyle)
	ax.legend()
	plt.tight_layout()
	plt.savefig(localfigoutput, dpi = 500)
	plt.show()	
	
if __name__ == "__main__":
	if prefix=="" or prefix==None:
		basename=os.path.basename(fastafilename)
		prefix=basename[:-(len(basename)-basename.rindex("."))] 
	outputname=GetWorkingBase(prefix) + ".predicted"	
	predictiondict={}
	if os.path.exists(outputname):
		predictiondict=LoadPrediction(outputname)
	ranklist=[]	
	if "," in classificationranks:
		ranklist=classificationranks.split(",")
	elif classificationranks !="":
		ranklist.append(classificationranks)
	higherranklist=[]	
	if "," in higherclassificationranks:
		higherranklist=higherclassificationranks.split(",")
	elif higherclassificationranks !="":
		higherranklist.append(higherclassificationranks)		
	if len(ranklist)==0:
		for rank in predictiondict.keys():
			prediction_datasets=predictiondict[rank]
			for datasetname in prediction_datasets.keys():
				datasetdict=prediction_datasets[datasetname]
				if "fasta filename" in datasetdict.keys() and "classification filename" in datasetdict.keys():
					if fastafilename== datasetdict["fasta filename"] and classificationfilename==datasetdict["classification filename"]:
						if not (rank in ranklist):
							ranklist.append(rank)	
	#load positions of classification for prediction
	positionlist,higherpositionlist,isError=GetPositionList(classificationfilename,ranklist,higherranklist)
	if isError==True:
		sys.exit()	
	#load sequences
	seqrecords=SeqIO.to_dict(SeqIO.parse(fastafilename, "fasta"))
	if len(seqrecords) >100:
		sys.setrecursionlimit(len(seqrecords)*2)
	if simfilename=="" or simfilename==None:
		simfilename=GetWorkingBase(prefix) + ".sim"
		
	#load similarity matrix
	simmatrix={}
	#load or compute simmatrix
	if endthreshold >=threshold and endthreshold >0:
		print(simfilename)
		if os.path.exists(simfilename):
			#question=input("Do you want to load the similarity matrix from the file " + simfilename + " (yes/no)?")
			#if question=="yes":
			print("Loading similarity matrix " + simfilename)
			simmatrix=LoadSim(simfilename)
		elif higherclassificationranks=="":	#predict globally
			print("Computing similarity matrix...")
			simmatrix=ComputeSim(fastafilename,seqrecords,mincoverage)
			print("Save similarity matrix " + simfilename)
			SaveSim(simmatrix,simfilename)	
	thresholdlist=[]
	intrathresholdlist=[]
	fmeasurelist=[]
	optthresholds=[]
	bestFmeasures=[]
	features=[]
	datasetnames=[]
	i=0
	#predicting
	for pos in positionlist:
		rank=ranklist[i]
		prediction_datasets={}
		if rank in predictiondict.keys():
			prediction_datasets=predictiondict[rank]
		datasets=GenerateDatasets(seqrecords,classificationfilename,pos,higherpositionlist,args.taxa)	
		if step==0 or (endthreshold < threshold) or (endthreshold==0 and threshold==0):
			#load existing prediction
			for datasetname in prediction_datasets.keys():
				#print(datasetname)
				if not datasetname in datasets.keys():
					continue 
				datasetdict=prediction_datasets[datasetname]
				if not ("fasta filename" in datasetdict.keys() and "classification filename" in datasetdict.keys()):
					continue
				elif fastafilename != datasetdict["fasta filename"] and classificationfilename!=datasetdict["classification filename"]:
					continue
				#seqno=datasetdict['sequence number']
				#groupno=datasetdict['group number']		
				thresholds,fmeasures,optthreshold,bestFmeasure,seqno,groupno,minalignmentlength=LoadPredictionAtPos(datasetdict)
				if not (groupno < minGroupNo or seqno < minSeqNo or (minalignmentlength >0 and minalignmentlength < mincoverage)):	#for visualization
					if (len(higherpositionlist)==0 and datasetname == "All") or (len(higherpositionlist)!=0):	
						#print(datasetname)
						thresholdlist.append(thresholds)
						fmeasurelist.append(fmeasures)
						optthresholds.append(optthreshold)
						bestFmeasures.append(bestFmeasure)
						features.append(rank)
						datasetnames.append(datasetname + "(" + str(seqno) + "," + str(groupno) + ")")			
		else:
			if os.path.exists(GetWorkingBase((os.path.basename(args.input))) + ".predict.log"):		
				os.system("rm " + GetWorkingBase((os.path.basename(args.input))) + ".predict.log")
			#datasets=GenerateDatasets(seqrecords,classificationfilename,pos,higherclassificationpos)
			for datasetname in datasets.keys():
				records=datasets[datasetname]
				seqno=len(records)
				#load classification at given positions 	
				classes,classification=LoadClasses(records.keys(),classificationfilename,pos)
#				#only predict when the numbers of the groups > 1
				if len(classes) < 2:
					continue
				datasetdict={}
				if datasetname in prediction_datasets.keys():
					datasetdict = prediction_datasets[datasetname]
				thresholds=[]
				fmeasures=[]		
				optthreshold=0
				bestFmeasure=0
				groupno=len(classes)
				print("Predicting optimal threshold to separate sequences at the " + rank + " level for " + datasetname)
				thresholds,fmeasures,optthreshold,bestFmeasure,isError=Predict(datasetname,datasetdict,records,classes,simmatrix)	
				if isError==False:
					datasetdict['min alignment length']=mincoverage
					datasetdict['fasta filename']=fastafilename
					datasetdict['classification filename']=classificationfilename
					if not (datasetname in prediction_datasets.keys()):
						prediction_datasets[datasetname]=datasetdict		
					if not (groupno < minGroupNo or seqno < minSeqNo):	#for visualization
						thresholdlist.append(thresholds)
						fmeasurelist.append(fmeasures)
						optthresholds.append(optthreshold)
						bestFmeasures.append(bestFmeasure)
						features.append(rank)
						datasetnames.append(datasetname + "(" + str(seqno) + "," + str(groupno) + ")")
			if not rank in predictiondict.keys():	
				predictiondict[rank] = prediction_datasets 
		i=i+1
	#save prediction and cut-offs for classification and identification
	if len(positionlist) >0:	
		if len(predictiondict.keys())>0:
			outputwithoutfmeasures=GetBase(outputname) + ".cutoffs.json"	
			SavePrediction(predictiondict,outputname,outputwithoutfmeasures)
			#SaveCutoffs(predictiondict,outputcutoffs)
			print("Only cutoffs for the clades with the numbers of sequences and subclades greater than  " + str(minSeqNo) + " and " + str(minGroupNo) + ", respectively. If you wish to predict also for clades with less numbers of sequences and groups, please reset minSeqNo and minGroupNo.")
			print("The prediction and cutoffs are saved in the files " + outputname + ", " + outputwithoutfmeasures + " and " + outputwithoutfmeasures + ".txt.")
	else:
		#load existing prediction for plotting
		for rank in predictiondict.keys():
			prediction_datasets=predictiondict[rank]
			for datasetname in prediction_datasets.keys():
				prediction_datasetname=prediction_datasets[datasetname]
				thresholds,fmeasures,optthreshold,bestFmeasure,feature,seqno,groupno,minalignmentlength=LoadPredictionAtPos(prediction_datasetname)
				if not (groupno < minGroupNo or seqno < minSeqNo or (minalignmentlength >0 and minalignmentlength < mincoverage)):	#for visualization
					thresholdlist.append(thresholds)
					fmeasurelist.append(fmeasures)
					optthresholds.append(optthreshold)
					bestFmeasures.append(bestFmeasure)
					features.append(feature)
					datasetnames.append(datasetname + "(" + str(seqno) + "," + str(groupno) + ")")
	#Plot the prediction results
	globalfigoutput=""
	localfigoutput=""
	if len(positionlist) == 1:
		globalfigoutput=GetBase(outputname) + "." + str(positionlist[0]) + ".global.png"
		localfigoutput=GetBase(outputname) + "." + str(positionlist[0]) + ".local.png"
	else: 	
		globalfigoutput=GetBase(outputname) + ".global.png"
		barplotfigoutput=GetBase(outputname) + ".local.png"
	if len(higherpositionlist)==0 or len(thresholdlist)==1:
		if len(thresholdlist) >0:
			#plot all predictions
			if label=="":
				label=prefix
			PlotPrediction(label,thresholdlist,fmeasurelist,optthresholds,bestFmeasures,features,datasetnames,globalfigoutput)	
			print("The prediction plot is saved in the file " + globalfigoutput + ".")
		else:
			print("Please check the parameters.")
	else:	
		if len(optthresholds) >0:
			#barplot the prediction results only
			if label=="":
				label=prefix
			PlotResults(label,optthresholds,bestFmeasures,features,datasetnames,localfigoutput)
			print("The prediction plot is saved in the file " + localfigoutput + ".")	
		else:
			print("Please check the parameters.")	
	if os.path.exists(GetWorkingBase((os.path.basename(args.input))) + ".predict.log"):		
		print("Please check the file " + GetWorkingBase((os.path.basename(args.input))) + ".predict.log for the prediction.")		
	
	
